import copy
import logging
import multiprocessing as mp
import os
import time
import warnings
from abc import ABC, abstractmethod
from collections import Counter
from functools import wraps
from os import path
from typing import Dict, Optional, Union, Tuple, List

import numpy as np

from autoconf import conf, cached_property
from autofit import exc
from autofit.database.sqlalchemy_ import sa
from autofit.graphical import (
    MeanField,
    AnalysisFactor,
    _HierarchicalFactor,
    FactorApproximation,
)
from autofit.graphical.utils import Status
from autofit.mapper.prior_model.collection import Collection
from autofit.non_linear.initializer import Initializer
from autofit.non_linear.parallel import SneakyPool
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.paths.directory import DirectoryPaths
from autofit.non_linear.paths.sub_directory_paths import SubDirectoryPaths
from autofit.non_linear.result import Result
from autofit.non_linear.timer import Timer
from .analysis import Analysis
from .analysis.combined import CombinedResult
from .analysis.indexed import IndexCollectionAnalysis
from .paths.null import NullPaths
from ..graphical.declarative.abstract import PriorFactor
from ..graphical.expectation_propagation import AbstractFactorOptimiser

logger = logging.getLogger(__name__)


def check_cores(func):
    """
    Checks how many cores the search has been configured to
    use and then returns None instead of calling the pool
    creation function in the case that only one core has
    been set.

    Parameters
    ----------
    func
        A function that creates a pool

    Returns
    -------
    None or a pool
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        self.logger.info(f"number_of_cores == {self.number_of_cores}...")
        if self.number_of_cores == 1:
            self.logger.info("...not using pool")
            return None
        return func(self, *args, **kwargs)

    return wrapper


class NonLinearSearch(AbstractFactorOptimiser, ABC):
    def __init__(
        self,
        name: Optional[str] = None,
        path_prefix: Optional[str] = None,
        unique_tag: Optional[str] = None,
        prior_passer: "PriorPasser" = None,
        initializer: Initializer = None,
        iterations_per_update: int = None,
        number_of_cores: int = 1,
        session: Optional[sa.orm.Session] = None,
        **kwargs,
    ):
        """
        Abstract base class for non-linear searches.L

        This class sets up the file structure for the non-linear search, which are standardized across all non-linear
        searches.

        Parameters
    ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        session
            An SQLAlchemy session instance so the results of the model-fit are written to an SQLite database.
        """
        super().__init__()

        from autofit.non_linear.paths.database import DatabasePaths

        path_prefix = path_prefix or ""

        self.path_prefix_no_unique_tag = path_prefix

        self._logger = None

        logger.info(f"Creating search")

        if unique_tag is not None:
            path_prefix = path.join(path_prefix, unique_tag)

        self.unique_tag = unique_tag

        if session is not None:
            logger.debug("Session found. Using database.")
            paths = DatabasePaths(
                name=name,
                path_prefix=path_prefix,
                session=session,
                save_all_samples=kwargs.get("save_all_samples", False),
                unique_tag=unique_tag,
            )
        elif name is not None or path_prefix:
            logger.debug("Session not found. Using directory output.")
            paths = DirectoryPaths(
                name=name, path_prefix=path_prefix, unique_tag=unique_tag
            )
        else:
            paths = NullPaths()

        self.paths: AbstractPaths = paths

        self.prior_passer = prior_passer or PriorPasser.from_config(config=self._config)

        self.force_pickle_overwrite = conf.instance["general"]["output"][
            "force_pickle_overwrite"
        ]

        if initializer is None:
            self.logger.debug("Creating initializer ")
            self.initializer = Initializer.from_config(config=self._config)
        else:
            self.initializer = initializer

        self.iterations_per_update = iterations_per_update or self._config(
            "updates", "iterations_per_update"
        )

        if conf.instance["general"]["hpc"]["hpc_mode"]:
            self.iterations_per_update = conf.instance["general"]["hpc"][
                "iterations_per_update"
            ]

        self.remove_state_files_at_end = self._config(
            "updates", "remove_state_files_at_end",
        )

        self.iterations = 0

        self.should_profile = conf.instance["general"]["profiling"]["should_profile"]

        self.silence = self._config("printing", "silence")

        if conf.instance["general"]["hpc"]["hpc_mode"]:
            self.silence = True

        self.kwargs = kwargs

        for key, value in self.config_dict_search.items():
            setattr(self, key, value)

        try:
            for key, value in self.config_dict_run.items():
                setattr(self, key, value)
        except KeyError:
            pass

        self.number_of_cores = number_of_cores

        if number_of_cores > 1 and any(
            os.environ.get(key) != "1"
            for key in (
                "OPENBLAS_NUM_THREADS",
                "MKL_NUM_THREADS",
                "OMP_NUM_THREADS",
                "VECLIB_MAXIMUM_THREADS",
                "NUMEXPR_NUM_THREADS",
            )
        ):

            warnings.warn(
                exc.SearchWarning(
                    """
                    The non-linear search is using multiprocessing (number_of_cores>1). 
                    
                    However, the following environment variables have not been set to 1:
                    
                    OPENBLAS_NUM_THREADS
                    MKL_NUM_THREADS
                    OMP_NUM_THREADS
                    VECLIB_MAXIMUM_THREADS
                    NUMEXPR_NUM_THREADS
                    
                    This can lead to performance issues, because both the non-linear search and libraries that may be
                    used in your `log_likelihood_function` evaluation (e.g. NumPy, SciPy, scikit-learn) may attempt to
                    parallelize over all cores available.
                    
                    This will lead to slow-down, due to overallocation of tasks over the CPUs.
                    
                    To mitigate this, set the environment variables to 1 via the following command on your
                    bash terminal / command line:
                    
                    export OPENBLAS_NUM_THREADS=1
                    export MKL_NUM_THREADS=1
                    export OMP_NUM_THREADS=1
                    export VECLIB_MAXIMUM_THREADS=1
                    export NUMEXPR_NUM_THREADS=1
                 
                    This means only the non-linear search is parallelized over multiple cores.
                    
                    If you "know what you are doing" and do not want these environment variables to be set to one, you 
                    can disable this warning by changing the following entry in the config files:
                    
                    `config -> general.yaml -> parallel: -> warn_environment_variable=False`
                    """
                )
            )

        self.optimisation_counter = Counter()

    __identifier_fields__ = tuple()

    def optimise(
        self, factor_approx: FactorApproximation, status: Status = Status(),
    ) -> Tuple[MeanField, Status]:
        """
        Perform optimisation for expectation propagation. Currently only
        applicable for ModelFactors created by the declarative interface.

        1. Analysis and model classes are extracted from the factor.
        2. Priors are updated from the mean field.
        3. Analysis and model are fit as usual.
        4. A new mean field is constructed with the (posterior) 'linking' priors.
        5. Projection is performed to produce an updated EPMeanField object.

        Output directories are generated according to the factor and the number
        of the search. For example a factor called "factor" would output:

        factor/optimization_0/<identifier>
        factor/optimization_1/<identifier>
        factor/optimization_2/<identifier>

        For the first, second and third optimizations respectively.

        Parameters
        ----------
        factor_approx
            A collection of messages defining the current best approximation to
            some global model
        status

        Returns
        -------
        An updated approximation to the model having performed optimisation on
        a single factor.
        """

        factor = factor_approx.factor

        _ = status
        if not isinstance(factor, (AnalysisFactor, PriorFactor, _HierarchicalFactor)):
            raise NotImplementedError(
                f"Optimizer {self.__class__.__name__} can only be applied to"
                f" AnalysisFactors, HierarchicalFactors and PriorFactors"
            )

        model = factor.prior_model.mapper_from_prior_arguments(
            {
                prior: prior.with_message(message)
                for prior, message in factor_approx.cavity_dist.arguments.items()
            }
        )

        analysis = factor.analysis

        number = self.optimisation_counter[factor.name]

        self.optimisation_counter[factor.name] += 1

        self.paths = SubDirectoryPaths(
            parent=self.paths,
            analysis_name=f"{factor.name}/optimization_{number}",
            is_flat=True,
        )

        result = self.fit(model=model, analysis=analysis)

        new_model_dist = MeanField.from_priors(result.projected_model.priors)

        status.result = result

        return new_model_dist, status

    @property
    def name(self):
        return self.paths.name

    def __getstate__(self):
        """
        Remove the logger for pickling
        """
        state = self.__dict__.copy()
        if "_logger" in state:
            del state["_logger"]
        if "paths" in state:
            del state["paths"]
        return state

    @property
    def logger(self):
        if not hasattr(self, "_logger"):
            self._logger = None
        if self._logger is None:
            logger_ = logging.getLogger(self.name)
            self._logger = logger_
        return self._logger

    @property
    def timer(self):
        return Timer(self.paths.samples_path)

    @property
    def paths(self) -> Optional[AbstractPaths]:
        return self._paths

    @paths.setter
    def paths(self, paths: Optional[AbstractPaths]):
        if paths is not None:
            paths.search = self
        self._paths = paths

    def copy_with_paths(self, paths):
        self.logger.debug(f"Creating a copy of {self._paths.name}")
        search_instance = copy.copy(self)
        search_instance.paths = paths
        search_instance._logger = None

        return search_instance

    class Fitness:
        def __init__(
            self, paths, model, analysis, samples_from_model, log_likelihood_cap=None
        ):

            self.i = 0

            self.paths = paths
            self.analysis = analysis

            self.model = model
            self.samples_from_model = samples_from_model

            self.log_likelihood_cap = log_likelihood_cap

        def __call__(self, parameters, *kwargs):

            try:
                figure_of_merit = self.figure_of_merit_from(parameter_list=parameters)

                if np.isnan(figure_of_merit):
                    return self.resample_figure_of_merit

                return figure_of_merit

            except exc.FitException:
                return self.resample_figure_of_merit

        def fit_instance(self, instance):
            log_likelihood = self.analysis.log_likelihood_function(instance=instance)

            if self.log_likelihood_cap is not None:
                if log_likelihood > self.log_likelihood_cap:
                    log_likelihood = self.log_likelihood_cap

            return log_likelihood

        def log_likelihood_from(self, parameter_list):

            instance = self.model.instance_from_vector(vector=parameter_list)
            log_likelihood = self.fit_instance(instance)

            return log_likelihood

        def log_posterior_from(self, parameter_list):

            log_likelihood = self.log_likelihood_from(parameter_list=parameter_list)
            log_prior_list = self.model.log_prior_list_from_vector(
                vector=parameter_list
            )

            return log_likelihood + sum(log_prior_list)

        def figure_of_merit_from(self, parameter_list):
            """
            The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space. This varies
            between different `NonLinearSearch`s, for example:

                - The *Optimizer* *PySwarms* uses the chi-squared value, which is the -2.0*log_posterior.
                - The *MCMC* algorithm *Emcee* uses the log posterior.
                - Nested samplers such as *Dynesty* use the log likelihood.
            """
            raise NotImplementedError()

        @staticmethod
        def prior(cube, model):

            # NEVER EVER REFACTOR THIS LINE! Haha.

            phys_cube = model.vector_from_unit_vector(unit_vector=cube)

            for i in range(len(phys_cube)):
                cube[i] = phys_cube[i]

            return cube

        @staticmethod
        def fitness(cube, model, fitness_function):
            return fitness_function(instance=model.instance_from_vector(cube))

        @property
        def samples(self):
            return self.samples_from_model(model=self.model)

        @property
        def resample_figure_of_merit(self):
            """
            If a sample raises a FitException, this value is returned to signify that the point requires resampling or
             should be given a likelihood so low that it is discard.
             """
            return -np.inf

    def fit_sequential(
        self,
        model,
        analysis: IndexCollectionAnalysis,
        info=None,
        pickle_files=None,
        log_likelihood_cap=None,
    ) -> CombinedResult:
        """
        Fit multiple analyses contained within the analysis sequentially.

        This can be useful for avoiding very high dimensional parameter spaces.

        Parameters
        ----------
        log_likelihood_cap
        analysis
            Multiple analyses that are fit sequentially
        model
            An object that represents possible instances of some model with a
            given dimensionality which is the number of free dimensions of the
            model.
        info
            Optional dictionary containing information about the fit that can be loaded by the aggregator.
        pickle_files : [str]
            Optional list of strings specifying the path and filename of .pickle files, that are copied to each
            model-fits pickles folder so they are accessible via the Aggregator.

        Returns
        -------
        An object combining the results of each individual optimisation.

        Raises
        ------
        AssertionError
            If the model has 0 dimensions.
        ValueError
            If the analysis is not a combined analysis
        """
        results = []

        _paths = self.paths
        original_name = self.paths.name or "analysis"

        model = analysis.modify_model(model=model)

        try:
            if not isinstance(model, Collection):
                model = [model for _ in range(len(analysis.analyses))]
        except AttributeError:
            raise ValueError(
                f"Analysis with type {type(analysis)} is not supported by fit_sequential"
            )

        for i, (model, analysis) in enumerate(zip(model, analysis.analyses)):
            self.paths = copy.copy(_paths)
            self.paths.name = f"{original_name}/{i}"
            results.append(
                self.fit(
                    model=model,
                    analysis=analysis,
                    info=info,
                    pickle_files=pickle_files,
                    log_likelihood_cap=log_likelihood_cap,
                )
            )
        self.paths = _paths
        return CombinedResult(results)

    def fit(
        self,
        model,
        analysis: "Analysis",
        info=None,
        pickle_files=None,
        log_likelihood_cap=None,
        bypass_nuclear_if_on : bool = False
    ) -> Union["Result", List["Result"]]:
        """
        Fit a model, M with some function f that takes instances of the
        class represented by model M and gives a score for their fitness.

        A model which represents possible instances with some dimensionality is fit.

        The analysis provides two functions. One visualises an instance of a model and the
        other scores an instance based on how well it fits some data. The search
        produces instances of the model by picking points in an N dimensional space.

        Parameters
        ----------
        log_likelihood_cap
        analysis : af.Analysis
            An object that encapsulates the data and a log likelihood function.
        model : ModelMapper
            An object that represents possible instances of some model with a
            given dimensionality which is the number of free dimensions of the
            model.
        info : dict
            Optional dictionary containing information about the fit that can be loaded by the aggregator.
        pickle_files : [str]
            Optional list of strings specifying the path and filename of .pickle files, that are copied to each
            model-fits pickles folder so they are accessible via the Aggregator.
        bypass_nuclear_if_on
            If nuclear mode is on (environment variable "PYAUTOFIT_NUCLEAR_MODE=1") passing this as True will
            bypass it.

        Returns
        -------
        An object encapsulating how well the model fit the data, the best fit instance
        and an updated model with free parameters updated to represent beliefs
        produced by this fit.

        Raises
        ------
        AssertionError
            If the model has 0 dimensions.
        """
        self.check_model(model=model)

        self.logger.info("Starting search")

        model = analysis.modify_model(model)
        self.paths.model = model
        self.paths.unique_tag = self.unique_tag
        self.paths.restore()

        analysis = analysis.modify_before_fit(paths=self.paths, model=model,)

        if not self.paths.is_complete or self.force_pickle_overwrite:
            self.logger.info("Saving path info")

            self.paths.save_all(
                search_config_dict=self.config_dict_search,
                info=info,
                pickle_files=pickle_files,
            )
            analysis.save_attributes_for_aggregator(paths=self.paths)

        if not self.paths.is_complete:
            self.logger.info("Not complete. Starting non-linear search.")

            self.timer.start()

            model.freeze()
            self._fit(
                model=model, analysis=analysis, log_likelihood_cap=log_likelihood_cap
            )
            model.unfreeze()

            self.paths.completed()

            samples = self.perform_update(
                model=model, analysis=analysis, during_analysis=False
            )

            result = analysis.make_result(
                samples=samples,
                model=model,
                sigma=self.prior_passer.sigma,
                use_errors=self.prior_passer.use_errors,
                use_widths=self.prior_passer.use_widths,
            )

            analysis.save_results_for_aggregator(paths=self.paths, result=result)

            self.paths.save_object("samples", samples)

        else:
            self.logger.info(f"Already completed, skipping non-linear search.")

            samples = self.paths.load_object("samples")

            result = analysis.make_result(
                samples=samples,
                model=model,
                sigma=self.prior_passer.sigma,
                use_errors=self.prior_passer.use_errors,
                use_widths=self.prior_passer.use_widths,
            )

            if self.force_pickle_overwrite:
                self.logger.info("Forcing pickle overwrite")
                self.paths.save_object("samples", samples)
                try:
                    self.paths.save_object("results", samples.results)
                except AttributeError:
                    self.paths.save_object("results", samples.results_internal)

                analysis.save_results_for_aggregator(paths=self.paths, result=result)

        self.paths.samples_to_csv(samples=samples)

        analysis = analysis.modify_after_fit(
            paths=self.paths, model=model, result=result
        )

        self.logger.info("Removing zip file")
        self.paths.zip_remove()

        if not bypass_nuclear_if_on:
            self.paths.zip_remove_nuclear()

        return result

    @abstractmethod
    def _fit(self, model, analysis, log_likelihood_cap=None):
        pass

    def check_model(self, model):
        if model is not None and model.prior_count == 0:
            raise AssertionError("Model has no priors! Cannot fit a 0 dimension model.")

    def config_dict_with_test_mode_settings_from(self, config_dict: Dict) -> Dict:
        return config_dict

    @property
    def _class_config(self) -> Dict:
        return self.config_type[self.__class__.__name__]

    @cached_property
    def config_dict_search(self) -> Dict:

        config_dict = copy.deepcopy(self._class_config["search"])

        for key, value in config_dict.items():
            try:
                config_dict[key] = self.kwargs[key]
            except KeyError:
                pass

        return config_dict

    @cached_property
    def config_dict_run(self) -> Dict:

        config_dict = copy.deepcopy(self._class_config["run"])

        for key, value in config_dict.items():
            try:
                config_dict[key] = self.kwargs[key]
            except KeyError:
                pass

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":

            logger.warning(f"TEST MODE ON: SEARCH WILL SKIP SAMPLING\n\n")

            config_dict = self.config_dict_with_test_mode_settings_from(
                config_dict=config_dict
            )

        return config_dict

    @property
    def config_dict_settings(self) -> Dict:
        return self._class_config["settings"]

    @property
    def config_type(self):
        raise NotImplementedError()

    def _config(self, section, attribute_name):
        """
        Get a config field from this search's section in non_linear.ini by a key and value type.

        Parameters
        ----------
        attribute_name: str
            The analysis_path of the field

        Returns
        -------
        attribute
            An attribute for the key with the specified type.
        """
        return self._class_config[section][attribute_name]

    def perform_update(self, model, analysis, during_analysis):
        """
        Perform an update of the `NonLinearSearch` results, which occurs every *iterations_per_update* of the
        non-linear search. The update performs the following tasks:

        1) Visualize the maximum log likelihood model.
        2) Output the model results to the model.reults file.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.
        during_analysis
            If the update is during a non-linear search, in which case tasks are only performed after a certain number
             of updates and only a subset of visualization may be performed.
        """

        self.iterations += self.iterations_per_update
        self.logger.info(
            f"{self.iterations} Iterations: Performing update (Visualization, outputting samples, etc.)."
        )

        self.timer.update()

        samples = self.samples_from(model=model)

        self.paths.save_object("samples", samples)

        if not isinstance(self.paths, NullPaths):
            self.plot_results(samples=samples)

        try:
            instance = samples.max_log_likelihood()
        except exc.FitException:
            return samples

        self.logger.debug("Visualizing")
        analysis.visualize(
            paths=self.paths, instance=instance, during_analysis=during_analysis
        )

        if self.should_profile:
            self.logger.debug("Profiling Maximum Likelihood Model")
            analysis.profile_log_likelihood_function(
                paths=self.paths, instance=instance,
            )

        self.logger.debug("Outputting model result")
        try:

            start = time.time()
            analysis.log_likelihood_function(instance=instance)
            log_likelihood_function_time = time.time() - start

            self.paths.save_summary(
                samples=samples,
                log_likelihood_function_time=log_likelihood_function_time,
            )
        except exc.FitException:
            pass

        if not during_analysis and self.remove_state_files_at_end:
            self.logger.debug("Removing state files")
            try:
                self.remove_state_files()
            except FileNotFoundError:
                pass

        return samples

    @property
    def samples_cls(self):
        raise NotImplementedError()

    def remove_state_files(self):
        pass

    def samples_from(self, model):
        raise NotImplementedError()

    @check_cores
    def make_pool(self):
        """Make the pool instance used to parallelize a `NonLinearSearch` alongside a set of unique ids for every
        process in the pool. If the specified number of cores is 1, a pool instance is not made and None is returned.

        The pool cannot be set as an attribute of the class itself because this prevents pickling, thus it is generated
        via this function before calling the non-linear search.

        The pool instance is also set up with a list of unique pool ids, which are used during model-fitting to
        identify a 'master core' (the one whose id value is lowest) which handles model result output, visualization,
        etc."""
        self.logger.info("...using pool")
        return mp.Pool(processes=self.number_of_cores)

    @check_cores
    def make_sneaky_pool(self, fitness_function: Fitness) -> Optional[SneakyPool]:
        """
        Create a pool for multiprocessing that uses slight-of-hand
        to avoid copying the fitness function between processes
        multiple times.

        Parameters
        ----------
        fitness_function
            An instance of a fitness class used to evaluate the
            likelihood that a particular model is correct

        Returns
        -------
        An implementation of a multiprocessing pool
        """
        self.logger.warning(
            "...using SneakyPool. This copies the likelihood function "
            "to each process on instantiation to avoid copying multiple "
            "times."
        )
        return SneakyPool(
            processes=self.number_of_cores, paths=self.paths, fitness=fitness_function
        )

    def __eq__(self, other):
        return isinstance(other, NonLinearSearch) and self.__dict__ == other.__dict__

    def plot_results(self, samples):
        pass


class PriorPasser:
    def __init__(self, sigma, use_errors, use_widths):
        """
        Class to package the API for prior passing.

        This class contains the parameters that controls how priors are passed from the results of one non-linear
        search to the next.

        Using the Phase API, we can pass priors from the result of one search to another follows:

        model_component.parameter = search1_result.model.model_component.parameter

        By invoking the 'model' attribute, the prior is passed following 3 rules:

        1) The new parameter uses a GaussianPrior. A ``GaussianPrior`` is ideal, as the 1D pdf results we compute at
           the end of a search are easily summarized as a Gaussian.

        2) The mean of the GaussianPrior is the median PDF value of the parameter estimated in search 1.

        This ensures that the initial sampling of the new search's non-linear starts by searching the region of
        non-linear parameter space that correspond to highest log likelihood solutions in the previous search.
        Thus, we're setting our priors to look in the 'correct' regions of parameter space.

        3) The sigma of the Gaussian will use the maximum of two values:

        (i) the 1D error of the parameter computed at an input sigma value (default sigma=3.0).
        (ii) The value specified for the profile in the 'config/priors/*.json' config
        file's 'width_modifer' field (check these files out now).

        The idea here is simple. We want a value of sigma that gives a GaussianPrior wide enough to search a
        broad region of parameter space, so that the model can change if a better solution is nearby. However,
        we want it to be narrow enough that we don't search too much of parameter space, as this will be slow or
        risk leading us into an incorrect solution! A natural choice is the errors of the parameter from the
        previous search.

        Unfortunately, this doesn't always work. Modeling can be prone to an effect called 'over-fitting' where
        we underestimate the parameter errors. This is especially true when we take the shortcuts in early
        searches, fast `NonLinearSearch` settings, simplified models, etc.

        Therefore, the 'width_modifier' in the json config files are our fallback. If the error on a parameter
        is suspiciously small, we instead use the value specified in the widths file. These values are chosen
        based on our experience as being a good balance broadly sampling parameter space but not being so narrow
        important solutions are missed.

        There are two ways a value is specified using the priors/width file:

        1) Absolute: In this case, the error assumed on the parameter is the value given in the config file. For
           example, if for the width on the parameter of a model component the width modifier reads "Absolute" with
           a value 0.05. This means if the error on the parameter was less than 0.05 in the previous search, the
           sigma of its GaussianPrior in this search will be 0.05.

        2) Relative: In this case, the error assumed on the parameter is the % of the value of the estimate value
           given in the config file. For example, if the parameter estimated in the previous search was 2.0, and the
           relative error in the config file reads "Relative" with a value 0.5, then the sigma of the GaussianPrior
           will be 50% of this value, i.e. sigma = 0.5 * 2.0 = 1.0.

        The PriorPasser allows us to customize at what sigma the error values the model results are computed at to
        compute the passed sigma values and customizes whether the widths in the config file, these computed errors,
        or both, are used to set the sigma values of the passed priors.

        The default values of the PriorPasser are found in the config file of every non-linear search, in the
        [prior_passer] section. All non-linear searches by default use a sigma value of 3.0, use_width=True and
        use_errors=True. We anticipate you should not need to change these values to get lens modeling to work
        proficiently!

        Example:

        Lets say in search 1 we fit a model, and we estimate that a parameter is equal to 4.0 +- 2.0, where the error
        value of 2.0 was computed at 3.0 sigma confidence. To pass this as a prior to search 2, we would write:

        model_component.parameter = result_1.model.model_component.parameter

        The prior on the parameter in search 2 would thus be a GaussianPrior, with mean=4.0 and
        sigma=2.0. If we had used a sigma value of 1.0 to compute the error, which reduced the estimate from 4.0 +- 2.0
        to 4.0 +- 0.5, the sigma of the Gaussian prior would instead be 0.5.

        If the error on the parameter in search 1 had been really small, lets say, 0.01, we would instead use the value
        of the parameter width in the priors config file to set sigma instead. Lets imagine the prior config file
        specifies that we use an "Absolute" value of 0.8 to link this prior. Then, the GaussianPrior in search 2 would
        have a mean=4.0 and sigma=0.8.

        If the prior config file had specified that we use an relative value of 0.8, the GaussianPrior in search 2 would
        have a mean=4.0 and sigma=3.2.
        """

        self.sigma = sigma
        self.use_errors = use_errors
        self.use_widths = use_widths

    @classmethod
    def from_config(cls, config):
        """
        Load the PriorPasser from a non_linear config file.
        """
        sigma = config("prior_passer", "sigma")
        use_errors = config("prior_passer", "use_errors")
        use_widths = config("prior_passer", "use_widths")
        return PriorPasser(sigma=sigma, use_errors=use_errors, use_widths=use_widths)



from autofit.plot.samples_plotters import SamplesPlotter
from autofit.non_linear.nest.dynesty.plotter import DynestyPlotter
from autofit.non_linear.nest.ultranest.plotter import UltraNestPlotter
from autofit.non_linear.mcmc.emcee.plotter import EmceePlotter
from autofit.non_linear.mcmc.zeus.plotter import ZeusPlotter
from autofit.non_linear.optimize.pyswarms.plotter import PySwarmsPlotter
from autofit.plot.output import Output


import copy
from os import path
import os
from typing import Optional

from autofit.database.sqlalchemy_ import sa

from autoconf import conf
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.nest import abstract_nest
from autofit.non_linear.nest.abstract_nest import AbstractNest
from autofit.non_linear.nest.ultranest.samples import SamplesUltraNest
from autofit.plot import UltraNestPlotter
from autofit.plot.output import Output


class UltraNest(abstract_nest.AbstractNest):
    __identifier_fields__ = (
        "draw_multiple",
        "ndraw_min",
        "ndraw_max",
        "min_num_live_points",
        "cluster_num_live_points",
        "insertion_test_zscore_threshold",
        "stepsampler_cls",
        "nsteps"
    )

    def __init__(
            self,
            name: str = "",
            path_prefix: str = "",
            unique_tag: Optional[str] = None,
            prior_passer: PriorPasser = None,
            iterations_per_update: int = None,
            number_of_cores: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        An UltraNest non-linear search.

        UltraNest is an optional requirement and must be installed manually via the command `pip install ultranest`.
        It is optional as it has certain dependencies which are generally straight forward to install (e.g. Cython).

        For a full description of UltraNest and its Python wrapper PyUltraNest, checkout its Github and documentation
        webpages:

        https://github.com/JohannesBuchner/UltraNest
        https://johannesbuchner.github.io/UltraNest/readme.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        iterations_per_update
            The number of iterations performed between every Dynesty back-up (via dumping the Dynesty instance as a
            pickle).
        number_of_cores
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        for key, value in self.config_dict_stepsampler.items():
            setattr(self, key, value)
            if self.config_dict_stepsampler["stepsampler_cls"] is None:
                self.nsteps = None

        self.logger.debug("Creating UltraNest Search")

    def config_dict_with_test_mode_settings_from(self, config_dict):

        return {
            **config_dict,
            "max_iters": 1,
            "max_ncalls": 1,
        }

    @property
    def config_dict_stepsampler(self):

        config_dict = copy.copy(self.config_type[self.__class__.__name__]["stepsampler"])

        for key, value in config_dict.items():
            try:
                config_dict[key] = self.kwargs[key]
            except KeyError:
                pass

        return config_dict

    @property
    def stepsampler(self):

        from ultranest import stepsampler

        config_dict_stepsampler = self.config_dict_stepsampler
        stepsampler_cls = config_dict_stepsampler["stepsampler_cls"]
        config_dict_stepsampler.pop("stepsampler_cls")

        if stepsampler_cls is None:
            return None
        elif stepsampler_cls == "RegionMHSampler":
            return stepsampler.RegionMHSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "AHARMSampler":
            config_dict_stepsampler.pop("scale")
            return stepsampler.AHARMSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "CubeMHSampler":
            return stepsampler.CubeMHSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "CubeSliceSampler":
            return stepsampler.CubeSliceSampler(**config_dict_stepsampler)
        elif stepsampler_cls == "RegionSliceSampler":
            return stepsampler.RegionSliceSampler(**config_dict_stepsampler)

    class Fitness(AbstractNest.Fitness):
        @property
        def resample_figure_of_merit(self):
            """If a sample raises a FitException, this value is returned to signify that the point requires resampling or
             should be given a likelihood so low that it is discard.

             -np.inf is an invalid sample value for Dynesty, so we instead use a large negative number."""
            return -1.0e99

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None):
        """
        Fit a model using Dynesty and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that includes the maximum log likelihood instance and full
        set of accepted ssamples of the fit.
        """

        try:
            import ultranest
        except ModuleNotFoundError:
            raise ModuleNotFoundError(
                "\n--------------------\n"
                "You are attempting to perform a model-fit using UltraNest. \n\n"
                "However, the optional library UltraNest (https://johannesbuchner.github.io/UltraNest/index.html) is "
                "not installed.\n\n"
                "Install it via the command `pip install ultranest==3.5.5`.\n\n"
                "----------------------"
            )

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis, log_likelihood_cap=log_likelihood_cap,
        )

        def prior_transform(cube):
            return model.vector_from_unit_vector(
                unit_vector=cube,
                ignore_prior_limits=True
            )

        sampler = ultranest.ReactiveNestedSampler(
            param_names=model.parameter_names,
            loglike=fitness_function.__call__,
            transform=prior_transform,
            log_dir=self.paths.samples_path,
            **self.config_dict_search
        )

        sampler.stepsampler = self.stepsampler

        finished = False

        while not finished:

            try:
                total_iterations = sampler.ncall
            except AttributeError:
                total_iterations = 0

            if self.config_dict_run["max_ncalls"] is not None:
                iterations = self.config_dict_run["max_ncalls"]
            else:
                iterations = total_iterations + self.iterations_per_update

            if iterations > 0:

                filter_list = ["max_ncalls", "dkl", "lepsilon"]
                config_dict_run = {
                    key: value for key, value
                    in self.config_dict_run.items()
                    if key
                    not in filter_list
                }

                config_dict_run["update_interval_ncall"] = iterations

                sampler.run(
                    max_ncalls=iterations,
                    **config_dict_run
                )

            self.paths.save_object(
                "results",
                sampler.results
            )

            self.perform_update(model=model, analysis=analysis, during_analysis=True)

            iterations_after_run = sampler.ncall

            if (
                    total_iterations == iterations_after_run
                    or iterations_after_run == self.config_dict_run["max_ncalls"]
            ):
                finished = True

    def samples_from(self, model: AbstractPriorModel):
        """
        Create a `Samples` object from this non-linear search's output files on the hard-disk and model.

        For MulitNest, this requires us to load:

            - The parameter samples, log likelihood values and weight_list from the ultranest.txt file.
            - The total number of samples (e.g. accepted + rejected) from resume.dat.
            - The log evidence of the model-fit from the ultranestsummary.txt file (if this is not yet estimated a
              value of -1.0e99 is used.

        Parameters
        ----------
        model
            The model which generates instances for different points in parameter space. This maps the points from unit
            cube values to physical values via the priors.
        """

        try:

            results_internal = self.paths.load_object(
                "results"
            )

        except FileNotFoundError:

            samples = self.paths.load_object(
                "samples"
            )
            results_internal = samples.results

        return SamplesUltraNest.from_results_internal(
            results_internal=results_internal,
            model=model,
            number_live_points=self.config_dict_run["min_num_live_points"],
            unconverged_sample_size=1,
            time=self.timer.time,
        )

    def plot_results(self, samples):

        if not samples.pdf_converged:
            return

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["ultranest"][name]

        plotter = UltraNestPlotter(
            samples=samples,
            output=Output(path=path.join(self.paths.image_path, "search"), format="png")
        )

        if should_plot("cornerplot"):
            plotter.cornerplot()

        if should_plot("runplot"):
            plotter.runplot()

        if should_plot("traceplot"):
            plotter.traceplot()



from autofit.mapper.model import ModelInstance
from autofit.mapper.model_mapper import ModelMapper
from autofit.non_linear.result import Result

from autofit.non_linear.mock.mock_samples import MockSamples


class MockResult(Result):
    def __init__(
            self,
            samples=None,
            instance=None,
            model=None,
            analysis=None,
            search=None,
    ):

        super().__init__(samples, model, search)

        self._instance = instance or ModelInstance()
        self.model = model or ModelMapper()
        self._samples = samples or MockSamples(max_log_likelihood_instance=self.instance)

        self.gaussian_tuples = None
        self.analysis = analysis
        self.search = search
        self.model = model

    def model_absolute(self, absolute):
        return self.model

    def model_relative(self, relative):
        return self.model

    @property
    def last(self):
        return self



class MockResultGrid(Result):
    def __init__(self, log_likelihood):
        # noinspection PyTypeChecker
        super().__init__(None, None)
        self._log_likelihood = log_likelihood
        self.model = log_likelihood

    @property
    def log_likelihood(self):
        return self._log_likelihood



import logging
import os
import shutil
from hashlib import md5
from pathlib import Path
from typing import Union

import yaml
from autofit.database.sqlalchemy_ import sa

from autofit.aggregator import Aggregator as ClassicAggregator
from autofit.database.aggregator import Aggregator as DatabaseAggregator
from autofit.non_linear.paths.database import DatabasePaths
from autofit.tools.util import zip_directory

logger = logging.getLogger(
    __name__
)


def update_identifiers_from_file(
        output_directory: Union[str, os.PathLike],
        map_filename: str
):
    """
    Update identifiers in a given directory by loading and modifying their
    identifier files.

    This is necessary when a change to source code means that pickles can no
    longer be loaded. Searches must also be re-run to replace pickles.

    Parameters
    ----------
    output_directory
        A directory containing output results
    map_filename
        Filename of a file mapping old fields to new
    """
    with open(map_filename) as f:
        field_map = yaml.safe_load(f)

    update_identifiers_from_dict(
        output_directory,
        field_map
    )


def update_identifiers_from_dict(
        output_directory: Union[str, os.PathLike],
        field_map: dict
):
    """
    Update identifiers in a given directory by loading and modifying their
    identifier files.

    This is necessary when a change to source code means that pickles can no
    longer be loaded. Searches must also be re-run to replace pickles.

    Parameters
    ----------
    output_directory
        A directory containing output results
    field_map
        A dictionary mapping old fields to new
    """
    aggregator = ClassicAggregator(
        output_directory
    )
    for output in aggregator:
        directory = output.directory

        def remove_source():
            shutil.rmtree(
                directory,
                ignore_errors=True
            )

        print(f"Processing {directory}")
        identifier_filename = f"{directory}/.identifier"
        with open(identifier_filename) as f:
            hash_list = f.read().split("\n")

        for i in range(len(hash_list)):
            string = hash_list[i]
            if string in field_map:
                new_value = field_map[string]
                print(f"Replacing {string} with {new_value}")
                hash_list[i] = new_value

        new_identifier = md5(".".join(
            hash_list
        ).encode("utf-8")).hexdigest()


        old_identifier = os.path.split(identifier_filename)[0]
        old_identifier = os.path.split(old_identifier)[-1]

     #   if new_identifier == output.search.paths.identifier:
        if new_identifier == old_identifier:
            logger.warning(
                f"Skipping {directory} as there is no change"
            )

            zip_file = f"{os.path.split(identifier_filename)[0]}.zip"

            if os.path.exists(zip_file):
                remove_source()
            continue

        new_directory = Path(
            directory
        ).parent / new_identifier

        print(
            f"Moving output from {directory} to {new_directory}"
        )

        os.makedirs(
            new_directory,
            exist_ok=True
        )

        for file in os.listdir(
                directory
        ):
            if file.endswith(
                    ".pickle"
            ):
                print(f"Skipping {file}")
                continue
            if not os.path.exists(
                    new_directory / file
            ):
                shutil.move(
                    f"{directory}/{file}",
                    new_directory
                )

        with open(
            new_directory / ".identifier",
            "w+"
        ) as f:
            f.write(
                "\n".join(
                    hash_list
                )
            )

        zip_directory(
            new_directory
        )

        shutil.rmtree(
            new_directory
        )
        remove_source()
        try:
            os.remove(
                f"{directory}.zip"
            )
        except FileNotFoundError:
            pass


def update_directory_identifiers(
        output_directory: Union[str, os.PathLike]
):
    """
    Update identifiers in a given directory.

    When identifiers were computed through an out of date method this
    can be used to move the data to a new directory with the correct
    identifier.

    search.pickle is replaced to ensure its internal identifier matches
    that of the directory.

    Parameters
    ----------
    output_directory
        A directory containing output results
    """
    aggregator = ClassicAggregator(
        output_directory
    )
    for output in aggregator:
        paths = output.search.paths
        try:
            os.remove(
                f"{paths.output_path}.zip"
            )
        except FileNotFoundError:
            pass

        source_directory = output.directory
        paths._identifier = None
        paths.save_identifier()
        target_directory = paths.output_path

        logger.info(
            f"Moving output from {source_directory} to {target_directory}"
        )

        os.makedirs(
            target_directory,
            exist_ok=True
        )

        for file in os.listdir(
                source_directory
        ):
            target_filename = f"{target_directory}/{file}"
            if not os.path.exists(
                    target_filename
            ):
                shutil.move(
                    f"{source_directory}/{file}",
                    target_filename
                )

        paths.save_object("search", output.search)

        shutil.rmtree(
            source_directory
        )

        paths.zip_remove()


def update_database_identifiers(
        session: sa.orm.Session
):
    """
    Update identifiers for a database.

    Parameters
    ----------
    session
        A SQLAlchemy session connected to the database
    """
    aggregator = DatabaseAggregator(session)

    args = list()

    for output in aggregator:
        search = output["search"]
        model = output["model"]
        paths = DatabasePaths(
            session=session,
            name=output.name,
            path_prefix=output.path_prefix,
            unique_tag=output.unique_tag,
        )
        paths.search = search
        paths.model = model

        args.append({
            "old_id": output.id,
            "new_id": paths.identifier
        })

    session.execute(
        "UPDATE fit SET id = :new_id WHERE id = :old_id",
        args
    )



import logging
from typing import Union, List

from autoconf import conf
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior.tuple_prior import TuplePrior
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.analysis.multiprocessing import AnalysisPool
from autofit.non_linear.paths.abstract import AbstractPaths
from autofit.non_linear.result import Result
from .analysis import Analysis

logger = logging.getLogger(__name__)


class CombinedResult:
    def __init__(self, results: List[Result]):
        """
        A `Result` object that is composed of multiple `Result` objects. This is used to combine the results of
        multiple `Analysis` objects into a single `Result` object, for example when performing a model-fitting
        analysis where there are multiple datasets.

        Parameters
        ----------
        results
            The list of `Result` objects that are combined into this `CombinedResult` object.
        """
        self.child_results = results

    def __getattr__(self, item: str):
        """
        Get an attribute of the first `Result` object in the list of `Result` objects.
        """
        return getattr(self.child_results[0], item)

    def __iter__(self):
        return iter(self.child_results)

    def __len__(self):
        return len(self.child_results)

    def __getitem__(self, item: int) -> Result:
        """
        Get a `Result` object from the list of `Result` objects.
        """
        return self.child_results[item]


class CombinedAnalysis(Analysis):
    def __new__(cls, *analyses, **kwargs):
        from .model_analysis import ModelAnalysis, CombinedModelAnalysis

        if any(isinstance(analysis, ModelAnalysis) for analysis in analyses):
            return object.__new__(CombinedModelAnalysis)
        return object.__new__(cls)

    def __init__(self, *analyses: Analysis):
        """
        Computes the summed log likelihood of multiple analyses
        applied to a single model.

        Either analyses are performed sequentially and summed,
        or they are mapped out to processes.

        If the number of cores is greater than one then the
        analyses are distributed across a number of processes
        equal to the number of cores.

        Parameters
        ----------
        analyses
        """
        self.analyses = analyses
        self._n_cores = None
        self._log_likelihood_function = None
        self.n_cores = conf.instance["general"]["analysis"]["n_cores"]

    def __getitem__(self, item):
        return self.analyses[item]

    def modify_before_fit(self, paths: AbstractPaths, model: AbstractPriorModel):
        """
        Modify the analysis before fitting.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        model
            The model which is to be fitted.
        """
        return CombinedAnalysis(
            *(analysis.modify_before_fit(paths, model) for analysis in self.analyses)
        )

    def modify_after_fit(
        self, paths: AbstractPaths, model: AbstractPriorModel, result: Result
    ):
        """
        Modify the analysis after fitting.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        model
            The model which is to be fitted.
        result
            The result of the fit.
        """
        return CombinedAnalysis(
            *(
                analysis.modify_after_fit(paths, model, result)
                for analysis in self.analyses
            )
        )

    @property
    def n_cores(self):
        return self._n_cores

    @n_cores.setter
    def n_cores(self, n_cores: int):
        """
        Set the number of cores this analysis should use.

        If the number of cores is greater than 1 then log likelihood
        computations are distributed across multiple processes.
        """
        self._n_cores = n_cores
        if self.n_cores > 1:
            analysis_pool = AnalysisPool(self.analyses, self.n_cores)
            self._log_likelihood_function = analysis_pool
        else:
            self._log_likelihood_function = self._summed_log_likelihood

    def _summed_log_likelihood(self, instance) -> float:
        """
        Compute a log likelihood by simply summing the log likelihood
        of each individual analysis computed for some instance.

        Parameters
        ----------
        instance
            An instance of a model

        Returns
        -------
        A combined log likelihood
        """
        return sum(
            analysis.log_likelihood_function(instance) for analysis in self.analyses
        )

    def log_likelihood_function(self, instance):
        return self._log_likelihood_function(instance)

    def _for_each_analysis(self, func, paths, *args):
        """
        Convenience function to call an underlying function for each
        analysis with a paths object with an integer attached to the
        end.

        Parameters
        ----------
        func
            Some function of the analysis class
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        """
        for (i, analysis), *args in zip(enumerate(self.analyses), *args):
            child_paths = paths.for_sub_analysis(analysis_name=f"analyses/analysis_{i}")
            func(child_paths, analysis, *args)

    def save_attributes_for_aggregator(self, paths: AbstractPaths):
        def func(child_paths, analysis):
            analysis.save_attributes_for_aggregator(child_paths,)

        self._for_each_analysis(func, paths)

    def save_results_for_aggregator(self, paths: AbstractPaths, result: Result):
        def func(child_paths, analysis, result_):
            analysis.save_results_for_aggregator(paths=child_paths, result=result_)

        self._for_each_analysis(func, paths, result)

    def visualize(self, paths: AbstractPaths, instance, during_analysis):
        """
        Visualise the instance according to each analysis.

        Visualisation output is distinguished by using an integer suffix
        for each analysis path.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        instance
            The maximum likelihood instance of the model so far in the non-linear search.
        during_analysis
            Is this visualisation during analysis?
        """

        def func(child_paths, analysis):
            analysis.visualize(child_paths, instance, during_analysis)

        self._for_each_analysis(func, paths)

    def profile_log_likelihood_function(
        self, paths: AbstractPaths, instance,
    ):
        """
        Profile the log likelihood function of the maximum likelihood model instance using each analysis.

        Profiling output is distinguished by using an integer suffix for each analysis path.

        Parameters
        ----------
        paths
            An object describing the paths for saving data (e.g. hard-disk directories or entries in sqlite database).
        instance
            The maximum likliehood instance of the model so far in the non-linear search.
        """

        def func(child_paths, analysis):
            analysis.profile_log_likelihood_function(
                child_paths, instance,
            )

        self._for_each_analysis(func, paths)

    def make_result(self, samples, model, sigma=1.0, use_errors=True, use_widths=False):
        child_results = [
            analysis.make_result(
                samples,
                model,
                sigma=sigma,
                use_errors=use_errors,
                use_widths=use_widths,
            )
            for analysis in self.analyses
        ]
        return CombinedResult(child_results)

    def __len__(self):
        return len(self.analyses)

    def __add__(self, other: Analysis):
        """
        Adding anything to a CombinedAnalysis results in another
        analysis containing all underlying analyses (no combined
        analysis children)

        Parameters
        ----------
        other
            Some analysis

        Returns
        -------
        An overarching analysis
        """
        if isinstance(other, CombinedAnalysis):
            return type(self)(*self.analyses, *other.analyses)
        return type(self)(*self.analyses, other)

    def with_free_parameters(
        self, *free_parameters: Union[Prior, TuplePrior, AbstractPriorModel]
    ):
        """
        Set some parameters as free parameters. The are priors which vary
        independently for each analysis in the collection.

        Parameters
        ----------
        free_parameters
            Parameters that are allowed to vary independently.

        Returns
        -------
        An analysis with freely varying parameters.
        """
        from .free_parameter import FreeParameterAnalysis

        return FreeParameterAnalysis(*self.analyses, free_parameters=free_parameters)



#!/usr/bin/env python
from sys import argv

import autofit as af

if __name__ == "__main__":
    aggregator = af.Aggregator(
        af.db.open_database(
            "sqlite://"
        )
    )
    aggregator.add_directory(
        argv[1]
    )
    print(aggregator)



import matplotlib.pyplot as plt
from functools import wraps
import logging
import os

from autofit.plot.output import Output

logger = logging.getLogger(__name__)

def log_value_error(func):

    @wraps(func)
    def wrapper(self, *args, **kwargs):

        try:
            return func(self, *args, **kwargs)
        except (ValueError, KeyError, AttributeError, AssertionError):
            self.log_plot_exception(func.__name__)

    return wrapper

def skip_plot_in_test_mode(func):
    """
    Skips visualization plots of non-linear searches if test mode is on.

    Parameters
    ----------
    func
        A function which plots a result of a non-linear search.

    Returns
    -------
        A function that plots a visual, or None if test mode is on.
    """

    @wraps(func)
    def wrapper(
        obj: object,
        *args,
        **kwargs
    ):
        """
        Skips visualization plots of non-linear searches if test mode is on.

        Parameters
        ----------
        obj
            An plotter object which performs visualization of a non-linear search.

        Returns
        -------
            A function that plots a visual, or None if test mode is on.
        """

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1":
            return

        return func(obj, *args, **kwargs)

    return wrapper



class SamplesPlotter:
    def __init__(
            self, 
            samples,
            output : Output = Output()
    ):

        self.samples = samples
        self.output = output

    @property
    def model(self):
        return self.samples.model

    def close(self):
        if plt.fignum_exists(num=1):
            plt.close()

    def log_plot_exception(self, plot_name : str):
        """
        Plotting the results of a ``dynesty`` model-fit before they have converged on an
        accurate estimate of the posterior can lead the ``dynesty`` plotting routines
        to raise a ``ValueError``.

        This exception is caught in each of the plotting methods below, and this
        function is used to log the behaviour.

        Parameters
        ----------
        plot_name
            The name of the ``dynesty`` plot which raised a ``ValueError``
        """

        logger.info(
            f"{self.__class__.__name__} unable to produce {plot_name} visual: posterior estimate therefore"
            "not yet sufficient for this model-fit is not yet robust enough to do this. Visual"
            "should be produced in later update, once posterior estimate is updated."
        )

class MCMCPlotter(SamplesPlotter):

    def _plot_trajectories(self, samples, log_posterior_list, **kwargs):

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7))

        for i in range(self.samples.model.prior_count):

            for walker_index in range(log_posterior_list.shape[1]):

                ax = axes[i]
                ax.plot(samples[:, walker_index, i], log_posterior_list[:, walker_index], alpha=0.3)

            ax.set_ylabel("Log Likelihood")
            ax.set_xlabel(self.model.parameter_labels_with_superscripts_latex[i])

        self.output.to_figure(structure=None, auto_filename="tracjectories")
        self.close()

    def _plot_likelihood_series(self, log_posterior_list, **kwargs):

        fig, axes = plt.subplots(1, figsize=(10, 7))

        for walker_index in range(log_posterior_list.shape[1]):

            axes.plot(log_posterior_list[:, walker_index], alpha=0.3)

        axes.set_ylabel("Log Likelihood")
        axes.set_xlabel("step number")

        self.output.to_figure(structure=None, auto_filename="likelihood_series")
        self.close()

    def _plot_time_series(self, samples, **kwargs):

        fig, axes = plt.subplots(self.samples.model.prior_count, figsize=(10, 7), sharex=True)

        for i in range(self.samples.model.prior_count):
            ax = axes[i]
            ax.plot(samples[:, :, i], alpha=0.3)
            ax.set_ylabel(self.model.parameter_labels_with_superscripts_latex[i])

        axes[-1].set_xlabel("step number")

        self.output.to_figure(structure=None, auto_filename="time_series")
        self.close()


from .ep_mean_field import EPMeanField
from .history import FactorHistory, EPHistory
from .optimiser import AbstractFactorOptimiser
from .optimiser import EPOptimiser
from .stochastic import StochasticEPOptimiser



from autofit.messages.normal import NormalMessage
from .abstract import Prior


class GaussianPrior(Prior):

    __identifier_fields__ = (
        "lower_limit",
        "upper_limit",
        "mean",
        "sigma"
    )

    def __init__(
        self,
        mean: float,
        sigma : float,
        lower_limit : float = float("-inf"),
        upper_limit : float = float("inf"),
    ):
        """
        A prior with a uniform distribution, defined between a lower limit and upper limit.

        The conversion of an input unit value, ``u``, to a physical value, ``p``, via the prior is as follows:

        .. math::

            p = \mu + (\sigma * sqrt(2) * erfcinv(2.0 * (1.0 - u))

        For example for ``prior = GaussianPrior(mean=1.0, sigma=2.0)``, an
        input ``prior.value_for(unit=0.5)`` is equal to 1.0.

        The mapping is performed using the message functionality, where a message represents the distirubtion
        of this prior.

        Parameters
        ----------
        mean
            The mean of the Gaussian distribution defining the prior.
        sigma
            The sigma value of the Gaussian distribution defining the prior.
        lower_limit
            A lower limit of the Gaussian distribution; physical values below this value are rejected.
        upper_limit
            A upper limit of the Gaussian distribution; physical values below this value are rejected.

        Examples
        --------

        prior = af.GaussianPrior(mean=1.0, sigma=2.0, lower_limit=0.0, upper_limit=2.0)

        physical_value = prior.value_for(unit=0.5)
        """
        super().__init__(
            lower_limit=lower_limit,
            upper_limit=upper_limit,
            message=NormalMessage(
                mean=mean,
                sigma=sigma,
                lower_limit=lower_limit,
                upper_limit=upper_limit,
            )
        )

    @classmethod
    def with_limits(
            cls,
            lower_limit: float,
            upper_limit: float
    ) -> "GaussianPrior":
        """
        Create a new gaussian prior centred between two limits
        with sigma distance between this limits.

        Note that these limits are not strict so exceptions will not
        be raised for values outside of the limits.

        This function is typically used in prior passing, where the
        result of a model-fit are used to create new Gaussian priors
        centred on the previously estimated median PDF model.

        Parameters
        ----------
        lower_limit
            The lower limit of the new Gaussian prior.
        upper_limit
            The upper limit of the new Gaussian Prior.

        Returns
        -------
        A new GaussianPrior
        """
        return cls(
            mean=(lower_limit + upper_limit) / 2,
            sigma=upper_limit - lower_limit,
        )

    def dict(self) -> dict:
        """
        A dictionary representation of this prior
        """
        prior_dict = super().dict()
        return {**prior_dict, "mean": self.mean, "sigma": self.sigma}

    @property
    def parameter_string(self) -> str:
        return f"mean = {self.mean}, sigma = {self.sigma}"



import logging
from abc import ABC, abstractmethod
from hashlib import md5
from typing import Union, Generator, Iterable, Optional

from .session_wrapper import SessionWrapper
from ..sqlalchemy_ import sa

logger = logging.getLogger(
    __name__
)


class Identifiable(ABC):
    @property
    @abstractmethod
    def id(self) -> str:
        """
        A unique identifier generated by hashing a string
        """

    def __eq__(
            self,
            other: Union["Identifiable", str]
    ) -> bool:
        """
        Compares ids
        """
        if isinstance(
                other,
                Identifiable
        ):
            return self.id == other.id
        if isinstance(
                other,
                str
        ):
            return self.id == other
        return False


class Step(Identifiable):
    def __init__(self, *strings: str):
        """
        A collection of SQL statements executed as one step
        in a database migration.

        Parameters
        ----------
        strings
            SQL statements
        """
        self.strings = strings

    @property
    def id(self) -> str:
        """
        Hash generated from underlying SQL statements
        """
        return md5(
            ":".join(
                self.strings
            ).encode(
                "utf-8"
            )
        ).hexdigest()

    def __str__(self):
        return "\n".join(self.strings)

    __repr__ = __str__


class Revision(Identifiable):
    def __init__(
            self,
            steps: Iterable[Step]
    ):
        """
        A specific revision of the database. This comprises
        a set of sequential steps and is uniquely identified
        by a hash on the hash of those steps.

        Parameters
        ----------
        steps
            Collections of SQL statements describing the changes
            between different versions of the database
        """
        self.steps = steps

    @property
    def id(self) -> str:
        """
        A unique identifier created by joining and hashing the
        identifiers of comprised steps.
        """
        return md5(
            ":".join(
                step.id for step
                in self.steps
            ).encode("utf-8")
        ).hexdigest()

    def __sub__(self, other: "Revision") -> "Revision":
        """
        Create a revision with steps that describe the difference
        between two revisions.

        For example, if the data base were at revision 2 and the
        code at revision 5, a 'revision' would be returned containing
        the steps required to migrate the database to revision 5.

        Parameters
        ----------
        other
            A previous revision

        Returns
        -------
        An object comprising steps required to move from the other
        revision to this revision.
        """
        return Revision(tuple(
            step for step in self.steps
            if step not in other.steps
        ))


class Migrator:
    def __init__(
            self,
            *steps: Step
    ):
        """
        Manages migration of an old database.

        The revision table is checked to see what version a database is on.
        This is compared to the identifier of the current revision to determine
        the set of Steps that must be executed to migrate the database.

        Parameters
        ----------
        steps
            All steps recorded for every migration
        """
        self._steps = steps

    @property
    def revisions(self) -> Generator[Revision, None, None]:
        """
        One revision exists for each sequential set of steps
        starting on the first step and terminating on any step
        """
        for i in range(1, len(self._steps) + 1):
            yield Revision(
                self._steps[:i]
            )

    def get_steps(
            self,
            revision_id: Optional[str] = None
    ) -> Iterable[Step]:
        """
        Retrieve steps required to go from the specified
        revision to the latest revision.

        Parameters
        ----------
        revision_id
            The identifier for a revision.

            If None or unrecognised then all steps are returned.

        Returns
        -------
        Steps required to get to the latest revision.
        """
        for revision in self.revisions:
            if revision_id == revision.id:
                return (self.latest_revision - revision).steps

        return self._steps

    @property
    def latest_revision(self) -> Revision:
        """
        The latest revision according to the steps passed to the
        Migrator
        """
        return Revision(
            self._steps
        )

    def migrate(self, session: sa.orm.Session):
        """
        Migrate the database that session points to to the current
        revision.

        Applies each required step and updates the revision identifier
        in the database.

        If no revision table is found then one is created.

        Parameters
        ----------
        session
            A session pointing at some database.
        """
        wrapper = SessionWrapper(
            session
        )
        revision_id = wrapper.revision_id
        steps = list(
            self.get_steps(
                revision_id
            )
        )
        if len(steps) == 0:
            logger.info(
                "Database already at latest revision"
            )
            return

        latest_revision_id = self.latest_revision.id

        logger.info(
            f"Performing migration from {revision_id} to {latest_revision_id} in {len(steps)} steps"
        )
        for step in steps:
            for string in step.strings:
                try:
                    session.execute(
                        string
                    )
                except sa.exc.OperationalError as e:
                    logger.debug(e)

        wrapper.revision_id = self.latest_revision.id

        logger.info(
            f"revision_id updated to {wrapper.revision_id}"
        )



import csv
import logging
import os
from copy import copy
from itertools import count
from pathlib import Path
from typing import List, Generator, Callable, ClassVar, Type, Union, Tuple

from autoconf import cached_property
from autofit.mapper.model import ModelInstance
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.abstract_search import NonLinearSearch
from autofit.non_linear.analysis import Analysis
from autofit.non_linear.grid.grid_search import make_lists
from autofit.non_linear.parallel import AbstractJob, Process, AbstractJobResult
from autofit.non_linear.result import Result
from autofit.text.text_util import padding


class JobResult(AbstractJobResult):
    def __init__(self, number: int, result: Result, perturbed_result: Result):
        """
        The result of a single sensitivity comparison

        Parameters
        ----------
        result
        perturbed_result
        """
        super().__init__(number)
        self.result = result
        self.perturbed_result = perturbed_result

    @property
    def log_likelihood_difference(self):
        return self.perturbed_result.log_likelihood - self.result.log_likelihood

    @property
    def log_likelihood_base(self):
        return self.result.log_likelihood

    @property
    def log_likelihood_perturbed(self):
        return self.perturbed_result.log_likelihood


class Job(AbstractJob):
    _number = count()

    use_instance = False

    def __init__(
        self,
        analysis_factory: "AnalysisFactory",
        model: AbstractPriorModel,
        perturbation_model: AbstractPriorModel,
        base_instance: ModelInstance,
        perturbation_instance: ModelInstance,
        search: NonLinearSearch,
        number: int,
    ):
        """
        Job to run non-linear searches comparing how well a model and a model with a perturbation
        fit the image.

        Parameters
        ----------
        model
            A base model that fits the image without a perturbation
        perturbation_model
            A model of the perturbation which has been added to the underlying image
        analysis_factory
            Factory to generate analysis classes which comprise a model and data
        search
            A non-linear search
        """
        super().__init__(number=number)

        self.analysis_factory = analysis_factory
        self.model = model

        self.perturbation_model = perturbation_model
        self.base_instance = base_instance
        self.perturbation_instance = perturbation_instance

        self.search = search.copy_with_paths(search.paths.for_sub_analysis("[base]",))
        self.perturbed_search = search.copy_with_paths(
            search.paths.for_sub_analysis("[perturbed]",)
        )

    @cached_property
    def analysis(self):
        return self.analysis_factory()

    def perform(self) -> JobResult:
        """
        - Create one model with a perturbation and another without
        - Fit each model against the perturbed image

        Returns
        -------
        An object comprising the results of the two fits
        """
        result = self.base_model_func()

        perturbed_model = copy(self.model)
        perturbed_model.perturbation = self.perturbation_model

        # TODO : This is what I added so that the Drawer runs use the correct subhalo model.

        # if self.use_instance:
        #     perturbed_model.perturbation = self.perturbation_instance

        perturbed_result = self.perturbation_model_func(perturbed_model=perturbed_model)
        return JobResult(
            number=self.number, result=result, perturbed_result=perturbed_result
        )

    def base_model_func(self):
        return self.search.fit(model=self.model, analysis=self.analysis)

    def perturbation_model_func(self, perturbed_model):
        return self.perturbed_search.fit(model=perturbed_model, analysis=self.analysis)


class SensitivityResult:
    def __init__(self, results: List[JobResult]):
        """
        The result of a sensitivity mapping

        Parameters
        ----------
        results
            The results of each sensitivity job
        """
        self.results = sorted(results)

    def __getitem__(self, item):
        return self.results[item]

    def __iter__(self):
        return iter(self.results)

    def __len__(self):
        return len(self.results)

    @property
    def log_likelihoods_base(self) -> List[float]:
        """
        The log likelihoods of the base model for each sensitivity fit
        """
        return [result.log_likelihood_base for result in self.results]

    @property
    def log_likelihoods_perturbed(self) -> List[float]:
        """
        The log likelihoods of the perturbed model for each sensitivity fit
        """
        return [result.log_likelihood_perturbed for result in self.results]

    @property
    def log_likelihood_differences(self) -> List[float]:
        """
        The log likelihood differences between the base and perturbed models
        """
        return [result.log_likelihood_difference for result in self.results]


class Sensitivity:
    def __init__(
        self,
        base_model: AbstractPriorModel,
        perturbation_model: AbstractPriorModel,
        simulation_instance,
        simulate_function: Callable,
        analysis_class: Type[Analysis],
        search: NonLinearSearch,
        job_cls: ClassVar = Job,
        number_of_steps: Union[Tuple[int], int] = 4,
        number_of_cores: int = 2,
        limit_scale: int = 1,
    ):
        """
        Perform sensitivity mapping to evaluate whether a perturbation
        can be detected if it occurs in different parts of an image.

        For a range from 0 to 1 with step_size, for each dimension of the
        perturbation_model, a perturbation is created and used in conjunction
        with the instance to create an image.

        For each of these images, a fit is run with just the model and with both
        the model and perturbation_model to compare how much better the image
        can be fit if the perturbation is included.

        Parameters
        ----------
        base_model
            A model that fits the instance well
        perturbation_model
            A model which provides a perturbations to be applied to the instance
            before creating images
        simulation_instance
            An instance of a model to which perturbations are applied prior to
            images being generated
        simulate_function
            A function that can convert an instance into an image
        analysis_class
            A class which can compare an image to an instance and evaluate fitness
        search
            A NonLinear search class which is copied and used to evaluate fitness
        number_of_cores
            How many cores does this computer have? Minimum 2.
        limit_scale
            Scales the priors for each perturbation model.
                A scale of 1 means priors have limits the same size as the grid square.
                A scale of 2 means priors have limits larger than the grid square with
                    width twice a grid square.
                A scale of 0.5 means priors have limits smaller than the grid square
                    with width half a grid square.
        """
        self.logger = logging.getLogger(f"Sensitivity ({search.name})")

        self.logger.info("Creating")

        self.instance = simulation_instance
        self.model = base_model

        self.search = search
        self.analysis_class = analysis_class

        self.perturbation_model = perturbation_model
        self.simulate_function = simulate_function

        self.job_cls = job_cls

        self.number_of_steps = number_of_steps
        self.number_of_cores = number_of_cores or 2

        self.limit_scale = limit_scale

    @property
    def step_size(self):
        """
        Returns
        -------
        step_size: float
            The size of a step in any given dimension in hyper space.
        """
        if isinstance(self.number_of_steps, tuple):
            return tuple(
                [1 / number_of_steps for number_of_steps in self.number_of_steps]
            )
        return 1 / self.number_of_steps

    def run(self) -> SensitivityResult:
        """
        Run fits and comparisons for all perturbations, returning
        a list of results.
        """
        self.logger.info("Running")

        self.search.paths.save_unique_tag(is_grid_search=True)

        headers = [
            "index",
            *self._headers,
            "log_likelihood_base",
            "log_likelihood_perturbed",
            "log_likelihood_difference",
        ]
        physical_values = list(self._physical_values)

        results = list()
        for result in Process.run_jobs(
            self._make_jobs(), number_of_cores=self.number_of_cores
        ):
            if isinstance(result, Exception):
                raise result

            results.append(result)
            results = sorted(results)

            os.makedirs(self.search.paths.output_path, exist_ok=True)
            with open(self.results_path, "w+") as f:
                writer = csv.writer(f)
                writer.writerow(headers)
                for result_ in results:
                    values = physical_values[result_.number]
                    writer.writerow(
                        padding(item)
                        for item in [
                            result_.number,
                            *values,
                            result_.log_likelihood_base,
                            result_.log_likelihood_perturbed,
                            result_.log_likelihood_difference,
                        ]
                    )

        result = SensitivityResult(results)

        self.search.paths.save_object("result", result)

        return SensitivityResult(results)

    @property
    def results_path(self):
        return Path(self.search.paths.output_path) / "results.csv"

    @property
    def _lists(self) -> List[List[float]]:
        """
        A list of hypercube vectors, used to instantiate
        the perturbation_model and create the individual
        perturbations.
        """
        return make_lists(self.perturbation_model.prior_count, step_size=self.step_size)

    @property
    def _physical_values(self) -> List[List[float]]:
        """
        Lists of physical values for each grid square
        """
        return [
            [
                prior.value_for(unit_value)
                for prior, unit_value in zip(
                    self.perturbation_model.priors_ordered_by_id, unit_values
                )
            ]
            for unit_values in self._lists
        ]

    @property
    def _headers(self) -> Generator[str, None, None]:
        """
        A name for each of the perturbed priors
        """
        for path, _ in self.perturbation_model.prior_tuples:
            yield path

    @property
    def _labels(self) -> Generator[str, None, None]:
        """
        One label for each perturbation, used to distinguish
        fits for each perturbation by placing them in separate
        directories.
        """
        for list_ in self._lists:
            strings = list()
            for value, prior_tuple in zip(list_, self.perturbation_model.prior_tuples):
                path, prior = prior_tuple
                value = prior.value_for(value)
                strings.append(f"{path}_{value}")
            yield "_".join(strings)

    @property
    def _perturbation_instances(self) -> Generator[ModelInstance, None, None]:
        """
        A list of instances each of which defines a perturbation to
        be applied to the image.
        """
        for list_ in self._lists:
            yield self.perturbation_model.instance_from_unit_vector(list_)

    @property
    def _perturbation_models(self) -> Generator[AbstractPriorModel, None, None]:
        """
        A list of models representing a perturbation at each grid square.

        By default models have priors with limits at the edges of a grid square.
        These limits can be scaled using the limit_scale variable. If the variable
        is 2 then the priors will have width twice the step size.
        """
        half_step = self.limit_scale * self.step_size / 2
        for list_ in self._lists:
            limits = [
                (
                    prior.value_for(max(0.0, centre - half_step)),
                    prior.value_for(min(1.0, centre + half_step)),
                )
                for centre, prior in zip(
                    list_, self.perturbation_model.priors_ordered_by_id
                )
            ]
            yield self.perturbation_model.with_limits(limits)

    @property
    def _searches(self) -> Generator[NonLinearSearch, None, None]:
        """
        A list of non-linear searches, each of which is applied to
        one perturbation.
        """
        for label in self._labels:
            yield self._search_instance(label)

    def _search_instance(self, name_path: str) -> NonLinearSearch:
        """
        Create a search instance, distinguished by its name

        Parameters
        ----------
        name_path
            A path to distinguish this search from other searches

        Returns
        -------
        A non linear search, copied from the instance search
        """
        paths = self.search.paths
        search_instance = self.search.copy_with_paths(
            paths.for_sub_analysis(name_path,)
        )

        return search_instance

    def _make_jobs(self) -> Generator[Job, None, None]:
        """
        Create a list of jobs to be run on separate processes.

        Each job fits a perturbed image with the original model
        and a model which includes a perturbation.
        """
        for number, (perturbation_instance, perturbation_model, search) in enumerate(
            zip(self._perturbation_instances, self._perturbation_models, self._searches)
        ):
            instance = copy(self.instance)
            instance.perturbation = perturbation_instance

            yield self.job_cls(
                analysis_factory=AnalysisFactory(
                    instance=instance,
                    simulate_function=self.simulate_function,
                    analysis_class=self.analysis_class,
                ),
                model=self.model,
                perturbation_model=perturbation_model,
                base_instance=self.instance,
                perturbation_instance=perturbation_instance,
                search=search,
                number=number,
            )


class AnalysisFactory:
    def __init__(
        self, instance, simulate_function, analysis_class,
    ):
        """
        Callable to delay simulation such that it is performed
        on the Job subprocess
        """
        self.instance = instance
        self.simulate_function = simulate_function
        self.analysis_class = analysis_class

    def __call__(self):
        dataset = self.simulate_function(self.instance)
        return self.analysis_class(dataset)



#!/usr/bin/env python

"""
Update the identifier for searches in a directory by mapping values in
the identifier file to new values.

This should be applied when the source code has changed meaning that pickles
cannot be loaded, for example when my.path.Class moves to my.Class.

Usage:
./update_identifiers_from_file.py directory map_file.yaml

map_file.yaml must be a yaml file mapping old values to new.

e.g.
old: new
something: else
"""

import logging
from sys import argv

from autofit.tools.update_identifiers import update_identifiers_from_file

if __name__ == "__main__":
    try:
        directory, map_file = argv[1:]
        update_identifiers_from_file(
            output_directory=directory,
            map_filename=map_file
        )
    except Exception as e:
        logging.exception(e)
        print(
            "Usage: update_identifiers_from_file.py directory map_file.yaml"
        )



import itertools
import os
import random
from abc import ABC, abstractmethod
from copy import copy
from typing import Union, Tuple

from autoconf import conf
from autofit import exc
from autofit.mapper.prior.arithmetic import ArithmeticMixin
from autofit.mapper.prior.deferred import DeferredArgument
from autofit.mapper.variable import Variable

epsilon = 1e-14


class Prior(Variable, ABC, ArithmeticMixin):
    __database_args__ = ("lower_limit", "upper_limit", "id_")

    _ids = itertools.count()

    def __init__(self, message, lower_limit=0.0, upper_limit=1.0, id_=None):
        """
        An object used to mappers a unit value to an attribute value for a specific
        class attribute.

        Parameters
        ----------
        lower_limit: Float
            The lowest value this prior can return
        upper_limit: Float
            The highest value this prior can return
        """
        if id_ is None:
            id_ = next(self._ids)
        super().__init__(id_=id_)
        self.message = message
        message.id_ = id_

        self.lower_limit = float(lower_limit)
        self.upper_limit = float(upper_limit)
        if self.lower_limit >= self.upper_limit:
            raise exc.PriorException(
                "The upper limit of a prior must be greater than its lower limit"
            )

        self.width_modifier = None

    @property
    def lower_unit_limit(self) -> float:
        """
        The lower limit for this prior in unit vector space
        """
        return self.unit_value_for(self.lower_limit)

    @property
    def upper_unit_limit(self) -> float:
        """
        The upper limit for this prior in unit vector space
        """
        return self.unit_value_for(self.upper_limit)

    def unit_value_for(self, physical_value: float) -> float:
        """
        Compute the unit value between 0 and 1 for the physical value.
        """
        return self.message.cdf(physical_value)

    def with_message(self, message):
        new = copy(self)
        new.message = message
        return new

    def new(self):
        """
        Returns a copy of this prior with a new id assigned making it distinct
        """
        new = copy(self)
        new.id = next(self._ids)
        return new

    def with_limits(self, lower_limit: float, upper_limit: float) -> "Prior":
        """
        Create a new instance of the same prior class with the passed limits.
        """
        new = self.__class__(
            lower_limit=max(lower_limit, self.lower_limit),
            upper_limit=min(upper_limit, self.upper_limit),
        )
        new.message = self.message
        return new

    @property
    def factor(self):
        """
        A callable PDF used as a factor in factor graphs
        """
        return self.message.factor

    def assert_within_limits(self, value):
        if (
            conf.instance["general"]["model"]["ignore_prior_limits"]
            or os.environ.get("PYAUTOFIT_TEST_MODE") == "1"
        ):
            return
        if not (self.lower_limit <= value <= self.upper_limit):
            raise exc.PriorLimitException(
                "The physical value {} for a prior "
                "was not within its limits {}, {}".format(
                    value, self.lower_limit, self.upper_limit
                )
            )

    @staticmethod
    def for_class_and_attribute_name(cls, attribute_name):
        prior_dict = conf.instance.prior_config.for_class_and_suffix_path(
            cls, [attribute_name]
        )
        return Prior.from_dict(prior_dict)

    @property
    def width(self):
        return self.upper_limit - self.lower_limit

    def random(self, lower_limit=0.0, upper_limit=1.0,) -> float:
        """
        A random value sampled from this prior
        """
        return self.value_for(
            random.uniform(
                max(lower_limit, self.lower_unit_limit),
                min(upper_limit, self.upper_unit_limit),
            )
        )

    def value_for(self, unit: float, ignore_prior_limits=False) -> float:
        """
        Return a physical value for a value between 0 and 1 with the transformation
        described by this prior.

        Parameters
        ----------
        unit
            A unit value between 0 and 1.

        Returns
        -------
        A physical value, mapped from the unit value accoridng to the prior.
        """
        result = self.message.value_for(unit)
        if not ignore_prior_limits:
            self.assert_within_limits(result)
        return result

    def instance_for_arguments(self, arguments):
        return arguments[self]

    def project(self, samples, weights):
        result = copy(self)
        result.message = self.message.project(
            samples=samples,
            log_weight_list=weights,
            id_=self.id,
            lower_limit=self.lower_limit,
            upper_limit=self.upper_limit,
        )
        return result

    def __getattr__(self, item):
        if item in ("__setstate__", "__getstate__"):
            raise AttributeError(item)
        return getattr(self.message, item)

    def __eq__(self, other):
        try:
            return self.id == other.id
        except AttributeError:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.id)

    def __repr__(self):
        return "<{} id={} lower_limit={} upper_limit={}>".format(
            self.__class__.__name__, self.id, self.lower_limit, self.upper_limit
        )

    def __str__(self):
        """The line of text describing this prior for the model_mapper.info file"""
        return f"{self.__class__.__name__} [{self.id}], {self.parameter_string}"

    @property
    @abstractmethod
    def parameter_string(self) -> str:
        pass

    def __float__(self):
        return self.value_for(0.5)

    @classmethod
    def from_dict(cls, prior_dict: dict) -> Union["Prior", DeferredArgument]:
        """
        Returns a prior from a JSON representation.

        Parameters
        ----------
        prior_dict : dict
            A dictionary representation of a prior including a type (e.g. Uniform) and all constructor arguments.

        Returns
        -------
        An instance of a child of this class.
        """
        if prior_dict["type"] == "Constant":
            return prior_dict["value"]
        if prior_dict["type"] == "Deferred":
            return DeferredArgument()

        from .uniform import UniformPrior
        from .log_uniform import LogUniformPrior
        from .gaussian import GaussianPrior
        from .log_gaussian import LogGaussianPrior

        prior_type_dict = {
            "Uniform": UniformPrior,
            "LogUniform": LogUniformPrior,
            "Gaussian": GaussianPrior,
            "LogGaussian": LogGaussianPrior,
        }

        # noinspection PyProtectedMember
        return prior_type_dict[prior_dict["type"]](
            **{
                key: value
                for key, value in prior_dict.items()
                if key not in ("type", "width_modifier", "gaussian_limits")
            }
        )

    def dict(self) -> dict:
        """
        A dictionary representation of this prior
        """
        prior_dict = {
            "lower_limit": self.lower_limit,
            "upper_limit": self.upper_limit,
            "type": self.name_of_class(),
        }
        return prior_dict

    @classmethod
    def name_of_class(cls) -> str:
        """
        A string name for the class, with the prior suffix removed.
        """
        return cls.__name__.replace("Prior", "")

    @property
    def limits(self) -> Tuple[float, float]:
        return self.lower_limit, self.upper_limit

    def gaussian_prior_model_for_arguments(self, arguments):
        return arguments[self]



import logging
from typing import Tuple

from autoconf import conf
from autofit.tools.util import open_

logger = logging.getLogger(__name__)


class FormatNode:
    def __init__(self):
        self._dict = dict()
        self.value = None

    def __getitem__(self, item):
        if item not in self._dict:
            self._dict[item] = FormatNode()
        return self._dict[item]

    def __len__(self):
        return len(self._dict)

    def items(self):
        return self._dict.items()


class TextFormatter:
    def __init__(self, line_length=90, indent=4):
        self.dict = FormatNode()
        self.line_length = line_length
        self.indent = indent

    def add_to_dict(self, path: Tuple[str, ...], value: str, info_dict: FormatNode):
        key = path[0]
        node = info_dict[key]
        if len(path) == 1:
            node.value = value
        else:
            self.add_to_dict(
                path[1:], value, node
            )

    def add(self, path: Tuple[str, ...], value):
        self.add_to_dict(path, value, self.dict)

    def dict_to_list(self, info_dict, line_length):
        lines = []
        for key, value in info_dict.items():
            indent_string = self.indent * " "
            if value.value is not None:
                value_string = str(value.value)
                space_string = max((line_length - len(key)), 1) * " "
                lines.append(f"{key}{space_string}{value_string}")

            if len(value) > 0:
                sub_lines = self.dict_to_list(
                    value, line_length=line_length - self.indent
                )
                if value.value is None:
                    lines.append(key)
                for line in sub_lines:
                    lines.append(f"{indent_string}{line}")
        return lines

    @property
    def list(self):
        return self.dict_to_list(self.dict, line_length=self.line_length)

    @property
    def text(self):
        return "\n".join(self.list)


def format_string_for_parameter_name(parameter_name: str) -> str:
    """
    Get the format for the label. Attempts to extract the key string associated with
    the dimension. Seems dodgy.

    Parameters
    ----------
    parameter_name
        A string label

    Returns
    -------
    format
        The format string (e.g {:.2f})
    """
    label_conf = conf.instance["notation"]["label_format"]

    try:
        # noinspection PyProtectedMember
        for key, value in sorted(
                label_conf["format"].items(),
                key=lambda item: len(item[0]),
                reverse=True,
        ):
            if key in parameter_name:
                return value
    except KeyError:
        pass

    logger.debug(
        "Could not find an entry for the parameter {} in the label_format.ini config at path {}".format(
            parameter_name, conf.instance.paths
        )
    )

    return "{:.4f}"


def convert_name_to_label(parameter_name, name_to_label):
    if not name_to_label:
        return parameter_name

    label_conf = conf.instance["notation"]["label"]

    try:
        return label_conf["label"][parameter_name]
    except KeyError:
        logger.debug(
            "Could not find an entry for the parameter {} in the label_format.iniconfig at paths {}".format(
                parameter_name, conf.instance.paths
            )
        )
        return parameter_name[0]


def add_whitespace(str0, str1, whitespace):
    return f"{str0}{str1.rjust(whitespace - len(str0) + len(str1))}"


def value_result_string_from(
        parameter_name, value, values_at_sigma=None, unit=None, format_string=None
):
    format_str = format_string or format_string_for_parameter_name(parameter_name)
    value = format_str.format(value)

    if unit is not None:
        unit = f" {unit}"
    else:
        unit = ""

    if values_at_sigma is None:
        return f"{value}{unit}"
    else:
        lower_value_at_sigma = format_str.format(values_at_sigma[0])
        upper_value_at_sigma = format_str.format(values_at_sigma[1])
        return f"{value} ({lower_value_at_sigma}, {upper_value_at_sigma}){unit}"


def parameter_result_latex_from(
        parameter_name,
        value,
        errors=None,
        superscript="",
        unit=None,
        format_string=None,
        name_to_label=False,
        include_name=True,
        include_quickmath=False,
):
    format_str = format_string or format_string_for_parameter_name(parameter_name)
    value = format_str.format(value)

    name = convert_name_to_label(
        parameter_name=parameter_name, name_to_label=name_to_label
    )

    if unit is not None:
        unit = f" {unit}"
    else:
        unit = ""

    if not superscript:
        superscript = ""
    else:
        superscript = f"^{{\\rm{{{superscript}}}}}"

    if errors is None:

        if include_name:
            parameter_result_latex = f"{name}{superscript} = {value}{unit}"
        else:
            parameter_result_latex = f"{value}{unit}"

    else:

        lower_value_at_sigma = format_str.format(errors[0])
        upper_value_at_sigma = format_str.format(errors[1])

        parameter_result = f"{value}^{{+{upper_value_at_sigma}}}_{{-{lower_value_at_sigma}}}{unit}"

        if include_name:
            parameter_result_latex = f"{name}{superscript} = {value}^{{+{upper_value_at_sigma}}}_{{-{lower_value_at_sigma}}}{unit}"
        else:
            parameter_result_latex = parameter_result

    if "e" in format_str:

        psplit = parameter_result.split("e")

        parameter_result_latex = (
            f"" 
            f"{psplit[0]}" 
            f"{psplit[1][3:]}"
            f"{psplit[2][3:]}" 
            f"{psplit[3][-1]}" 
            f" \\times 10^{{{int(psplit[1][1:3])}}}"
        )

    if not include_quickmath:
        return f"{parameter_result_latex} & "
    return f"${parameter_result_latex}$ & "


def output_list_of_strings_to_file(file, list_of_strings):
    with open_(file, "w") as f:
        f.write("".join(list_of_strings))



from autoconf.exc import PriorException


class MessageException(PriorException):
    """
    Raised when some assertion about the parameterization of a message is not met
    """


class PathsException(Exception):
    pass


class FitException(Exception):
    """
    An exception to be thrown if the non linear search must resample; equivalent to returning an infinitely bad fit
    """

    pass


class PriorLimitException(FitException, PriorException):
    pass


class PipelineException(Exception):
    pass


class DeferredInstanceException(Exception):
    """
    Exception raised when an attempt is made to access an attribute or function of a
    deferred instance prior to instantiation
    """

    pass


class AggregatorException(Exception):
    pass


class GridSearchException(Exception):
    pass


class HistoryException(Exception):
    """
    Thrown when insufficient factor history is present for a given operation
    """


class InitializerException(Exception):
    """
    Raises exceptions associated with the `non_linear.initializer` module and `Initializer` classes.

    For example if all initial samples have identical figures of merit.
    """


class SamplesException(Exception):
    pass


class SearchException(Exception):
    pass


class SamplesWarning(Warning):
    """
    Raises warnings associated with the `non_linear` module and `NonLinearSearch` classes.

    For example if the search is parallel but enviromental variables controlling multithreading are sub-optimal.
    """
    pass


class SearchWarning(Warning):
    pass


from os import path
import os
import matplotlib.pyplot as plt
import numpy as np

import autofit as af

"""
The `analysis.py` module contains the dataset and log likelihood function which given a model instance (set up by
the non-linear search) fits the dataset and returns the log likelihood of that model.
"""

class Analysis(af.Analysis):
    def __init__(self, data: np.ndarray, noise_map:np.ndarray):
        """
        In this example the `Analysis` object only contains the data and noise-map. It can be easily extended,
        for more complex data-sets and model fitting problems.

        Parameters
        ----------
        data
            A 1D numpy array containing the data (e.g. a noisy 1D Gaussian) fitted in the workspace examples.
        noise_map
            A 1D numpy array containing the noise values of the data, used for computing the goodness of fit
            metric.
        """
        super().__init__()

        self.data = data
        self.noise_map = noise_map

    def log_likelihood_function(self, instance: af.ModelInstance) -> float:
        """
        Determine the log likelihood of a fit of multiple profiles to the dataset.

        Parameters
        ----------
        instance : af.Collection
            The model instances of the profiles.

        Returns
        -------
        The log likelihood value indicating how well this model fit the dataset.
        """

        xvalues = np.arange(self.data.shape[0])
        model_data_1d = np.zeros(self.data.shape[0])

        try:
            for profile in instance:
                try:
                    model_data_1d += profile.model_data_1d_via_xvalues_from(xvalues=xvalues)
                except AttributeError:
                    pass
        except TypeError:
            model_data_1d += instance.model_data_1d_via_xvalues_from(xvalues=xvalues)

        residual_map = self.data - model_data_1d
        chi_squared_map = (residual_map / self.noise_map) ** 2.0
        log_likelihood = -0.5 * sum(chi_squared_map)

        return log_likelihood

    def visualize(self, paths: af.DirectoryPaths, instance: af.ModelInstance, during_analysis : bool):
        """
        During a model-fit, the `visualize` method is called throughout the non-linear search and is used to output
        images indicating the quality of the fit so far..

        The `instance` passed into the visualize method is maximum log likelihood solution obtained by the model-fit
        so far and it can be used to provide on-the-fly images showing how the model-fit is going.

        For your model-fitting problem this function will be overwritten with plotting functions specific to your
        problem.

        Parameters
        ----------
        paths
            The PyAutoFit paths object which manages all paths, e.g. where the non-linear search outputs are stored,
            visualization, and the pickled objects used by the aggregator output by this function.
        instance
            An instance of the model that is being fitted to the data by this analysis (whose parameters have been set
            via a non-linear search).
        during_analysis
            If True the visualization is being performed midway through the non-linear search before it is finished,
            which may change which images are output.
        """

        xvalues = np.arange(self.data.shape[0])
        model_data_1d = np.zeros(self.data.shape[0])

        try:
            for profile in instance:
                try:
                    model_data_1d += profile.model_data_1d_via_xvalues_from(xvalues=xvalues)
                except AttributeError:
                    pass
        except TypeError:
            model_data_1d += instance.model_data_1d_via_xvalues_from(xvalues=xvalues)

        plt.errorbar(
            x=xvalues,
            y=self.data,
            yerr=self.noise_map,
            color="k",
            ecolor="k",
            elinewidth=1,
            capsize=2,
        )
        plt.plot(range(self.data.shape[0]), model_data_1d, color="r")
        plt.title("Dynesty model fit to 1D Gaussian + Exponential dataset.")
        plt.xlabel("x values of profile")
        plt.ylabel("Profile normalization")

        os.makedirs(paths.image_path, exist_ok=True)
        plt.savefig(path.join(paths.image_path, "model_fit.png"))
        plt.clf()



from functools import wraps
from typing import Dict, Tuple, Set, Optional
import operator
from collections import ChainMap

import numpy as np
from scipy.linalg import block_diag

from autoconf import cached_property

from autofit.mapper.operator import (
    LinearOperator,
    IdentityOperator,
    DiagonalMatrix,
    CholeskyOperator,
    MatrixOperator,
    QROperator,
)
from autofit.mapper.variable import (
    Variable,
    VariableData,
    VariableLinearOperator,
    InverseVariableOperator,
    rtruediv,
    rmul,
)

from autofit.graphical.utils import FlattenArrays
from autofit.graphical.factor_graphs.abstract import FactorValue


def ldiv(op, val):
    return op.ldiv(val)


def _merged_operator_vdata(op):
    @wraps(op)
    def __op__(self, *args):
        out = VariableData()
        for op in self.operators:
            variables = op.variables
            subsets = (VariableData.subset(arg, variables) for arg in args)
            out.update(op(*subsets))

        return out


class MergedVariableOperator(VariableLinearOperator):
    def __init__(self, *operators: Tuple[VariableLinearOperator]):
        self.operators = operators

    __mul__ = _merged_operator_vdata(operator.mul)
    __rmul__ = _merged_operator_vdata(operator.mul)
    __rtruediv__ = _merged_operator_vdata(operator.truediv)
    ldiv = _merged_operator_vdata(ldiv)

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    def __getitem__(self, variable):
        for op in self.operators:
            if variable in op.variables:
                return op[variable]

        raise KeyError(f"{variable} not in operator")

    @property
    def variables(self):
        variables = self.operators[0].variables
        for op in self.operators[1:]:
            variables = variables | op.variables

        return variables

    @property
    def is_diagonal(self):
        return all(op.is_diagonal for op in self.operators.values())

    def to_block(self, cls=None):
        blocks = {}
        for op in self.operators:
            blocks.update(op.to_block(cls).operators)

        return VariableOperator(blocks)

    def to_full(self):
        full_ops = [op.to_full() for op in self.operators]
        full = block_diag(*(op.operator.to_dense() for op in full_ops))
        param_shapes = FlattenArrays(
            (v, s) for op in full_ops for v, s in op.param_shapes.items()
        )
        return VariableFullOperator(full_ops[0].operator.from_dense(full), param_shapes)

    def inv(self):
        return type(self)(*(op.inv() for op in self.operators))

    @cached_property
    def log_det(self):
        return sum(M.log_det for M in self.operators.values())

    def diagonal(self):
        diag = VariableData()
        for op in self.operators:
            diag.update(op.diagonal())

        return diag

    def update(self, *args):
        operators = list(self.operators)
        for u, v in args:
            operators = [op.update(u, v) for op in operators]

        return MergedVariableOperator(operators)

    def lowrankupdate(self, *values: VariableData) -> "VariableFullOperator":
        operators = list(self.operators)
        for u in values:
            operators = [op.lowrankupdate(u) for op in operators]

        return MergedVariableOperator(operators)

    def lowrankdowndate(self, values: VariableData) -> "VariableFullOperator":
        operators = list(self.operators)
        for u in values:
            operators = [op.lowrankdowndate(u) for op in operators]

        return MergedVariableOperator(operators)

    def diagonalupdate(self, values: VariableData) -> "VariableFullOperator":
        operators = [op.diagonalupdate(values) for op in self.operators]
        return MergedVariableOperator(operators)


def _variable_binary_op(op):
    @wraps(op)
    def __op__(self, other):
        if isinstance(other, FactorValue):
            other = other.to_dict()

        if isinstance(other, (dict, VariableData)):
            return VariableData(
                {
                    k: op(self.operators[k], other[k])
                    for k in self.operators.keys() & other.keys()
                }
            )
        elif isinstance(other, VariableOperator):
            return type(self)(
                {
                    v: op(self.operators[v], other.operators[v])
                    for v in self.operators.keys() & other.operators.keys()
                }
            )
        elif isinstance(other, VariableLinearOperator):
            return op(self.to_full(), other.to_full())
        else:
            return type(self)({k: op(val, other) for k, val in self.operators.items()})

    return __op__


class VariableOperator(VariableLinearOperator):
    """ """

    def __init__(self, operators: Dict[Variable, LinearOperator]):
        self.operators = operators

    __add__ = _variable_binary_op(operator.add)
    __sub__ = _variable_binary_op(operator.sub)
    __mul__ = _variable_binary_op(operator.mul)
    __rmul__ = _variable_binary_op(rmul)
    __truediv__ = _variable_binary_op(operator.truediv)
    __rtruediv__ = _variable_binary_op(rtruediv)
    ldiv = _variable_binary_op(ldiv)

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    def __getitem__(self, variable):
        return self.operators[variable]

    @property
    def is_diagonal(self):
        return all(op.is_diagonal for op in self.operators.values())

    @property
    def variables(self) -> Set[Variable]:
        return self.operators.keys()

    def inv(self):
        return type(self)({v: op.inv() for v, op in self.operators.items()})

    @cached_property
    def log_det(self):
        return sum(M.log_det for M in self.operators.values())

    @classmethod
    def from_scales(cls, scales):
        return cls({v: DiagonalMatrix(scale) for v, scale in scales.items()})

    @classmethod
    def from_covariances(cls, covs):
        return cls(
            {v: CholeskyOperator.from_dense(cov).inv() for v, cov in covs.items()}
        )

    @classmethod
    def from_pos_definite(cls, pos_defs):
        return cls({v: CholeskyOperator.from_dense(M) for v, M in pos_defs.items()})

    @classmethod
    def from_dense(cls, pos_defs):
        return cls({v: MatrixOperator(M) for v, M in pos_defs.items()})

    @classmethod
    def from_diagonal(cls, diag: VariableData) -> "VariableFullOperator":
        operators = {v: DiagonalMatrix(d) for v, d in diag.items()}
        return cls(operators)

    def diagonal(self):
        return VariableData({v: op.diagonal() for v, op in self.operators.items()})

    def to_diagonal(self) -> "VariableOperator":
        return type(self)({v: DiagonalMatrix(d) for v, d in self.diagonal().items()})

    def abs(self) -> "VariableOperator":
        return type(self)({v: d.matrixabs() for v, d in self.diagonal().items()})

    def blocks(self) -> VariableData:
        return VariableData({v: op.to_dense() for v, op in self.operators.items()})

    def to_block(self, cls=None) -> "VariableOperator":
        return self

    @cached_property
    def param_shapes(self):
        return FlattenArrays({v: op.lshape for v, op in self.operators.items()})

    def to_full(self) -> "VariableFullOperator":
        M = self.param_shapes.flatten2d(
            {v: op.to_dense() for v, op in self.operators.items()}
        )
        return VariableFullOperator(MatrixOperator(M), self.param_shapes)

    def update(self, *args: Tuple[VariableData, VariableData]):
        operators = self.operators.copy()
        for (u, v) in args:
            for k in operators.keys() & u.keys() & v.keys():
                operators[k] = operators[k].update(u[k], v[k])

        return type(self)(operators)

    def diagonalupdate(self, d: VariableData):
        return type(self)(
            {
                k: self.operators[k].diagonalupdate(d[k])
                for k in self.operators.keys() & d.keys()
            }
        )


def _variablefull_binary_op(op):
    @wraps(op)
    def __op__(self: "VariableFullOperator", other):
        if isinstance(other, FactorValue):
            other = other.to_dict()

        if isinstance(other, dict):
            return self.param_shapes.unflatten(
                op(self.operator, self.param_shapes.flatten(other))
            )
        if isinstance(other, VariableFullOperator):
            other = other.operator
        elif isinstance(other, VariableLinearOperator):
            other = other.to_full().operator

        op_new = op(self.operator, other)
        return type(self)(op_new, self.param_shapes)

    return __op__


class VariableFullOperator(VariableLinearOperator):

    __add__ = _variablefull_binary_op(operator.add)
    __sub__ = _variablefull_binary_op(operator.sub)
    __mul__ = _variablefull_binary_op(operator.mul)
    __rmul__ = _variablefull_binary_op(rmul)
    __truediv__ = _variablefull_binary_op(operator.truediv)
    __rtruediv__ = _variablefull_binary_op(rtruediv)
    ldiv = _variablefull_binary_op(ldiv)
    rdiv = __rtruediv__
    rmul = __rmul__
    mul = __mul__
    __matmul__ = __mul__

    def __init__(self, op: LinearOperator, param_shapes: FlattenArrays):
        self.operator = op
        self.param_shapes = param_shapes

    def __getitem__(self, variable):
        return self.param_shapes.extract(variable, self.operator).reshape(
            self.param_shapes[variable] * 2
        )

    def inv(self):
        return type(self)(self.operator.inv(), self.param_shapes)

    @property
    def is_diagonal(self):
        return self.operator.is_diagonal

    @property
    def variables(self) -> Set[Variable]:
        return self.param_shapes.keys()

    @classmethod
    def from_posdef(
        cls, M: np.ndarray, param_shapes: FlattenArrays
    ) -> "VariableFullOperator":
        return cls(CholeskyOperator.from_dense(M), param_shapes)

    @classmethod
    def from_dense(
        cls, M: np.ndarray, param_shapes: FlattenArrays
    ) -> "VariableFullOperator":
        return cls(MatrixOperator.from_dense(M), param_shapes)

    @classmethod
    def from_diagonal(cls, diag: VariableData) -> "VariableFullOperator":
        param_shapes = FlattenArrays.from_arrays(diag)
        operator = DiagonalMatrix(param_shapes.flatten(diag))
        return cls(operator, param_shapes)

    @classmethod
    def from_blocks(cls, Ms: VariableData) -> "VariableFullOperator":
        param_shapes = FlattenArrays(
            {v: np.shape(M)[: np.ndim(M) // 2] for v, M in Ms.items()}
        )
        return cls.from_dense(param_shapes.flatten2d(Ms), param_shapes)

    def blocks(self) -> VariableData:
        return self.param_shapes.unflatten2d(self.operator.to_dense())

    def to_block(self, cls=None) -> VariableOperator:
        blocks = self.blocks()
        cls = cls or type(self.operator)
        return VariableOperator({k: cls.from_dense(M) for k, M in blocks.items()})

    def diagonal(self) -> VariableData:
        return self.param_shapes.unflatten(self.operator.diagonal())

    def to_diagonal(self):
        diagonal = DiagonalMatrix(self.operator.diagonal())
        return type(self)(diagonal, self.param_shapes)

    def abs(self):
        matrixabs = self.operator.matrixabs()
        return type(self)(matrixabs, self.param_shapes)

    def to_full(self):
        return self

    @classmethod
    def from_optresult(cls, opt_result) -> "VariableFullOperator":
        param_shapes = opt_result.param_shapes

        cov = opt_result.result.hess_inv
        if not isinstance(cov, np.ndarray):
            # if optimiser is L-BFGS-B then convert
            # implicit hess_inv into dense matrix
            cov = cov.todense()

        return cls.from_dense(cov, param_shapes)

    @cached_property
    def log_det(self):
        return self.operator.log_det

    def update(self, *args):
        op = self.operator.update(
            *(
                (self.param_shapes.flatten(u), self.param_shapes.flatten(v))
                for u, v in args
            )
        )
        return type(self)(op, self.param_shapes)

    def lowrankupdate(self, values: VariableData) -> "VariableFullOperator":
        v = self.param_shapes.flatten(values)
        return type(self)(self.operator.lowrankupdate(v), self.param_shapes)

    def lowrankdowndate(self, values: VariableData) -> "VariableFullOperator":
        v = self.param_shapes.flatten(values)
        return type(self)(self.operator.lowrankdowndate(v), self.param_shapes)

    def diagonalupdate(self, values: VariableData) -> "VariableFullOperator":
        v = self.param_shapes.flatten(values)
        return type(self)(self.operator.diagonalupdate(v), self.param_shapes)


class IdentityVariableOperator(VariableLinearOperator):
    def __init__(self, variables=()):
        self._variables = set(variables)

    def _identity(self, values: VariableData) -> VariableData:
        return values

    def __getitem__(self, variable):
        if variable in self:
            return identity_operator
        else:
            raise KeyError(f"{variable} not present")

    @property
    def variables(self):
        return self._variables

    __mul__ = _identity
    __rtruediv__ = _identity
    __rmul__ = _identity
    ldiv = _identity
    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__
    quad = _identity
    invquad = _identity

    @property
    def log_det(self):
        return 0.0

    def update(self, *args):
        (u, v), *next_args = args
        param_shapes = FlattenArrays.from_arrays(u)
        uv = param_shapes.flatten(u)[:, None] * param_shapes.flatten(v)[None, :]
        out = VariableFullOperator.from_dense(uv, param_shapes)
        return out.update(*next_args)


identity_operator = IdentityOperator()
identity_variable_operator = IdentityVariableOperator()


def _rect_variable_binary_op(op, left=True):
    if left:
        op_ = op
    else:

        def op_(x, y):
            return op(y, x)

    @wraps(op)
    def __op__(self, other):
        if isinstance(other, FactorValue):
            other = other.to_dict()

        if isinstance(other, (dict, VariableData)):
            operators = self.operators if left else self.T.operators
            return VariableData(
                {
                    kl: sum(op_(right_op, other).values())
                    for kl, right_op in operators.items()
                }
            )
        elif isinstance(other, RectVariableOperator):
            lops, rops = self.operators, other.operators
            return type(self)(
                {
                    vl: {
                        vr: op_(lops[vl][vr], rops[vl][vr])
                        for vr in lops[vl].keys() & rops[vl]
                    }
                    for vl in lops.keys() & rops.keys()
                }
            )
        elif isinstance(other, VariableLinearOperator):
            raise NotImplementedError()
        else:
            return type(self)(
                {
                    k: {vl: op_(vop, other) for vl, vop in val.operators.items()}
                    for k, val in self.operators.items()
                }
            )

    return __op__


class RectVariableOperator(VariableLinearOperator):
    """ """

    def __init__(self, operators: Dict[Variable, Dict[Variable, LinearOperator]]):
        self.operators = {k: VariableOperator(ops) for k, ops in operators.items()}
        self.left_variables = operators.keys()

    @cached_property
    def right_variables(self):
        return ChainMap(*(vop.operators for vop in self.operators.values())).keys()

    @property
    def transposed_operators(self):
        transposed_operators = {v: {} for v in self.right_variables}
        for lv, rops in self.operators.items():
            for rv, op in rops.operators.items():
                transposed_operators[rv][lv] = op

        return transposed_operators

    @property
    def T(self):
        return type(self)(self.transposed_operators)

    __add__ = _rect_variable_binary_op(operator.add)
    __sub__ = _rect_variable_binary_op(operator.sub)
    __mul__ = _rect_variable_binary_op(operator.mul)
    __rmul__ = _rect_variable_binary_op(operator.mul, left=False)
    __truediv__ = _rect_variable_binary_op(operator.truediv)
    __rtruediv__ = _rect_variable_binary_op(operator.truediv, left=False)
    ldiv = _rect_variable_binary_op(ldiv)

    rdiv = __rtruediv__
    rmul = __rmul__
    lmul = __mul__
    __matmul__ = __mul__

    @property
    def is_diagonal(self):
        return False

    @property
    def variables(self) -> Set[Variable]:
        return self.left_variables.union(self.right_variables)

    def inv(self):
        raise NotImplementedError()

    @cached_property
    def log_det(self):
        raise NotImplementedError()

    @classmethod
    def from_scales(cls, scales):
        raise NotImplementedError()

    @classmethod
    def from_covariances(cls, covs):
        raise NotImplementedError()

    @classmethod
    def from_pos_definite(cls, pos_defs):
        raise NotImplementedError()

    @classmethod
    def from_dense(
        cls,
        data: Dict[Variable, Dict[Variable, np.ndarray]],
        *,
        var_shapes: Optional[Dict[Variable, Tuple[int, ...]]] = None,
        values: Optional[VariableData] = None
    ) -> "RectVariableOperator":
        if values:
            var_shapes = {v: np.shape(val) for v, val in values.items()}
        if var_shapes:
            return cls(
                {
                    vl: {
                        vr: MatrixOperator.from_dense(
                            M, var_shapes[vl] + var_shapes[vr], len(var_shapes[vl])
                        )
                        for vr, M in val.items()
                    }
                    for vl, val in data.items()
                }
            )
        else:
            return cls(
                {
                    vl: {vr: MatrixOperator.from_dense(M) for vr, M in val.items()}
                    for vl, val in data.items()
                }
            )

    @classmethod
    def from_diagonal(cls, diag: VariableData) -> "VariableFullOperator":
        raise NotImplementedError()

    def diagonal(self):
        raise NotImplementedError()

    def to_diagonal(self) -> "VariableOperator":
        raise NotImplementedError()

    def blocks(self) -> VariableData:
        return {
            vl: {vl: op for vr, op in rops.operators.items()}
            for vl, rops in self.operators.items()
        }

    def to_dense(self) -> Dict[Variable, VariableData]:
        return VariableData(
            {
                v0: VariableData(
                    {v1: op.to_dense() for v1, op in rops.operators.items()}
                )
                for v0, rops in self.operators.items()
            }
        )

    def to_block(self, cls=None) -> "VariableOperator":
        return self

    @cached_property
    def lparam_shapes(self):
        return FlattenArrays({v: op.lshape for v, op in self.operators.items()})

    def to_full(self) -> "VariableFullOperator":
        M = self.param_shapes.flatten2d(
            {v: op.to_dense() for v, op in self.operators.items()}
        )
        return VariableFullOperator(MatrixOperator(M), self.param_shapes)

    def update(self, *args: Tuple[VariableData, VariableData]):
        operators = self.operators.copy()
        for (u, v) in args:
            for k in operators.keys() & u.keys() & v.keys():
                operators[k] = operators[k].update(u[k], v[k])

        return type(self)(operators)

    def diagonalupdate(self, d: VariableData):
        return type(self)(
            {
                k: self.operators[k].diagonalupdate(d[k])
                for k in self.operators.keys() & d.keys()
            }
        )



from os import path
from typing import Optional

import numpy as np

from autoconf import conf
from autofit import exc
from autofit.database.sqlalchemy_ import sa
from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.abstract_search import PriorPasser
from autofit.non_linear.initializer import AbstractInitializer
from autofit.non_linear.optimize.abstract_optimize import AbstractOptimizer
from autofit.non_linear.optimize.pyswarms.samples import SamplesPySwarms
from autofit.plot import PySwarmsPlotter
from autofit.plot.output import Output


class AbstractPySwarms(AbstractOptimizer):
    def __init__(
            self,
            name: Optional[str] = None,
            path_prefix: Optional[str] = None,
            unique_tag: Optional[str] = None,
            prior_passer: Optional[PriorPasser] = None,
            initializer: Optional[AbstractInitializer] = None,
            iterations_per_update: int = None,
            number_of_cores: int = None,
            session: Optional[sa.orm.Session] = None,
            **kwargs
    ):
        """
        A PySwarms Particle Swarm Optimizer global non-linear search.

        For a full description of PySwarms, checkout its Github and readthedocs webpages:

        https://github.com/ljvmiranda921/pyswarms
        https://pyswarms.readthedocs.io/en/latest/index.html

        Parameters
        ----------
        name
            The name of the search, controlling the last folder results are output.
        path_prefix
            The path of folders prefixing the name folder where results are output.
        unique_tag
            The name of a unique tag for this model-fit, which will be given a unique entry in the sqlite database
            and also acts as the folder after the path prefix and before the search name.
        prior_passer
            Controls how priors are passed from the results of this `NonLinearSearch` to a subsequent non-linear search.
        initializer
            Generates the initialize samples of non-linear parameter space (see autofit.non_linear.initializer).
        number_of_cores
            The number of cores Emcee sampling is performed using a Python multiprocessing Pool instance. If 1, a
            pool instance is not created and the job runs in serial.
        session
            An SQLalchemy session instance so the results of the model-fit are written to an SQLite database.
        """

        number_of_cores = (
            self._config("parallel", "number_of_cores")
            if number_of_cores is None
            else number_of_cores
        )

        super().__init__(
            name=name,
            path_prefix=path_prefix,
            unique_tag=unique_tag,
            prior_passer=prior_passer,
            initializer=initializer,
            iterations_per_update=iterations_per_update,
            number_of_cores=number_of_cores,
            session=session,
            **kwargs
        )

        self.logger.debug("Creating PySwarms Search")

    class Fitness(AbstractOptimizer.Fitness):
        def __call__(self, parameters, *kwargs):

            figures_of_merit = []

            for params_of_particle in parameters:

                try:
                    figure_of_merit = self.figure_of_merit_from(
                        parameter_list=params_of_particle
                    )
                except exc.FitException:
                    figure_of_merit = -2.0 * self.resample_figure_of_merit

                if np.isnan(figure_of_merit):
                    figure_of_merit = -2.0 * self.resample_figure_of_merit

                figures_of_merit.append(figure_of_merit)

            return np.asarray(figures_of_merit)

        def figure_of_merit_from(self, parameter_list):
            """
            The figure of merit is the value that the `NonLinearSearch` uses to sample parameter space.

            PySwarms uses the chi-squared value, which is the -2.0*log_posterior.
            """
            return -2.0 * self.log_posterior_from(parameter_list=parameter_list)

    def _fit(self, model: AbstractPriorModel, analysis, log_likelihood_cap=None):
        """
        Fit a model using PySwarms and the Analysis class which contains the data and returns the log likelihood from
        instances of the model, which the `NonLinearSearch` seeks to maximize.

        Parameters
        ----------
        model : ModelMapper
            The model which generates instances for different points in parameter space.
        analysis : Analysis
            Contains the data and the log likelihood function which fits an instance of the model to the data, returning
            the log likelihood the `NonLinearSearch` maximizes.

        Returns
        -------
        A result object comprising the Samples object that inclues the maximum log likelihood instance and full
        chains used by the fit.
        """

        fitness_function = self.fitness_function_from_model_and_analysis(
            model=model, analysis=analysis
        )

        if self.paths.is_object("points"):

            init_pos = self.paths.load_object("points")[-1]
            total_iterations = self.paths.load_object("total_iterations")

            self.logger.info("Existing PySwarms samples found, resuming non-linear search.")

        else:

            unit_parameter_lists, parameter_lists, log_posterior_list = self.initializer.samples_from_model(
                total_points=self.config_dict_search["n_particles"],
                model=model,
                fitness_function=fitness_function,
            )

            init_pos = np.zeros(shape=(self.config_dict_search["n_particles"], model.prior_count))

            for index, parameters in enumerate(parameter_lists):

                init_pos[index, :] = np.asarray(parameters)

            total_iterations = 0

            self.logger.info("No PySwarms samples found, beginning new non-linear search. ")

        ## TODO : Use actual limits

        vector_lower = model.vector_from_unit_vector(
            unit_vector=[1e-6] * model.prior_count,
            ignore_prior_limits=True
        )
        vector_upper = model.vector_from_unit_vector(
            unit_vector=[0.9999999] * model.prior_count,
            ignore_prior_limits=True
        )

        lower_bounds = [lower for lower in vector_lower]
        upper_bounds = [upper for upper in vector_upper]

        bounds = (np.asarray(lower_bounds), np.asarray(upper_bounds))

        self.logger.info("Running PySwarmsGlobal Optimizer...")

        while total_iterations < self.config_dict_run["iters"]:

            pso = self.sampler_from(
                model=model,
                fitness_function=fitness_function,
                bounds=bounds,
                init_pos=init_pos
            )

            iterations_remaining = self.config_dict_run["iters"] - total_iterations

            iterations = min(self.iterations_per_update, iterations_remaining)

            if iterations > 0:

                pso.optimize(objective_func=fitness_function.__call__, iters=iterations)

                total_iterations += iterations

                self.paths.save_object(
                    "total_iterations",
                    total_iterations
                )
                self.paths.save_object(
                    "points",
                    pso.pos_history
                )
                self.paths.save_object(
                    "log_posterior_list",
                    [-0.5 * cost for cost in pso.cost_history]
                )

                self.perform_update(
                    model=model, analysis=analysis, during_analysis=True
                )

                init_pos = self.paths.load_object("points")[-1]

        self.logger.info("PySwarmsGlobal complete")

    def config_dict_with_test_mode_settings_from(self, config_dict):

        return {
            **config_dict,
            "iters": 1,
        }

    def fitness_function_from_model_and_analysis(self, model, analysis, log_likelihood_cap=None):

        return AbstractPySwarms.Fitness(
            paths=self.paths,
            model=model,
            analysis=analysis,
            samples_from_model=self.samples_from,
            log_likelihood_cap=log_likelihood_cap
        )

    def sampler_from(self, model, fitness_function, bounds, init_pos):
        raise NotImplementedError()

    def samples_from(self, model):

        return SamplesPySwarms.from_results_internal(
            results_internal=self.paths.load_object("points"),
            model=model,
            log_posterior_list=self.paths.load_object("log_posterior_list"),
            total_iterations=self.paths.load_object("total_iterations"),
            time=self.timer.time
        )

    def plot_results(self, samples):

        def should_plot(name):
            return conf.instance["visualize"]["plots_search"]["pyswarms"][name]

        plotter = PySwarmsPlotter(
            samples=samples,
            output=Output(path=path.join(self.paths.image_path, "search"), format="png")
        )

        if should_plot("contour"):
            plotter.contour()

        if should_plot("cost_history"):
            plotter.cost_history()

        if should_plot("trajectories"):
            plotter.trajectories()

        if should_plot("time_series"):
            plotter.time_series()



from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.mapper.prior_model.collection import Collection
from .analysis import Analysis
from .indexed import IndexCollectionAnalysis


class ModelAnalysis(Analysis):
    def __init__(self, analysis: Analysis, model: AbstractPriorModel):
        """
        Comprises a model and an analysis that can be applied to instances of that model.

        Parameters
        ----------
        analysis
        model
        """
        self.analysis = analysis
        self.model = model

    def __getattr__(self, item):
        return getattr(self.analysis, item)

    def log_likelihood_function(self, instance):
        return self.analysis.log_likelihood_function(instance)

    def make_result(self, samples, model, sigma=1.0, use_errors=True, use_widths=False):
        """
        Return the correct type of result by calling the underlying analysis.
        """
        return self.analysis.make_result(
            samples=samples,
            model=model,
            sigma=sigma,
            use_errors=use_errors,
            use_widths=use_widths,
        )


class CombinedModelAnalysis(IndexCollectionAnalysis):
    def modify_model(self, model: AbstractPriorModel) -> Collection:
        """
        Creates a collection with one model for each analysis. For each ModelAnalysis
        the model is used; for other analyses the default model is used.

        Parameters
        ----------
        model
            A default model

        Returns
        -------
        A collection of models, one for each analysis.
        """
        return Collection(
            [
                analysis.modify_model(analysis.analysis.model)
                if isinstance(analysis.analysis, ModelAnalysis)
                else analysis.modify_model(model)
                for analysis in self.analyses
            ]
        )



import logging
import math
from itertools import count
from multiprocessing import Process
from multiprocessing import Queue

from autofit.non_linear.parallel.sneaky import StopCommand

logger = logging.getLogger(
    __name__
)


class AnalysisProcess(Process):
    _id = count()

    def __init__(
            self,
            analyses
    ):
        """
        A process that performs one or more analyses on each
        instance passed to it

        Parameters
        ----------
        analyses
            A list of analyses this process performs
        """
        super().__init__()
        self.analyses = analyses
        self.instance_queue = Queue()
        self.queue = Queue()

        self.name = f"analysis_process_{next(self._id)}"

        self.logger = logging.getLogger(
            self.name
        )
        self.logger.info("Created")

    def _run(self):
        """
        Private run function permitting process to be stopped using a return
        when a StopCommand is passed
        """
        while True:
            instance = self.instance_queue.get()
            if instance is StopCommand:
                return
            for analysis in self.analyses:
                try:
                    self.queue.put(
                        analysis.log_likelihood_function(
                            instance
                        )
                    )
                except Exception as e:
                    self.queue.put(e)

    def run(self):
        """
        Run this process, completing each job in the job_queue and
        passing the result to the queue.
        """

        self.logger.debug("starting")

        self._run()

        self.logger.debug("terminating")
        self.queue.close()


class AnalysisPool:
    def __init__(
            self,
            analyses,
            n_cores: int
    ):
        """
        A pool which distributes analysis evenly across
        n_cores processes.

        Callable as a log-likelihood function.

        Parameters
        ----------
        analyses
            A list of analyses to be performed in parallel
        n_cores
            The number of cores available to perform analyses over
        """
        self.analyses = analyses
        self.n_cores = n_cores

        self.n_analyses = len(analyses)
        n_processes = min(
            self.n_analyses,
            n_cores
        )

        analyses_per_process = math.ceil(
            self.n_analyses / n_processes
        )

        self.processes = list()

        for n in range(
                n_processes
        ):
            analyses_ = analyses[
                        n * analyses_per_process: (n + 1) * analyses_per_process
                        ]
            process = AnalysisProcess(
                analyses=analyses_
            )

            self.processes.append(
                process
            )
            process.start()

    def __del__(self):
        """
        Called when the map goes out of scope.

        Tell each process to terminate with a StopCommand and then join
        each process with a timeout of one second.
        """
        logger.debug(
            "Deconstructing SneakyMap"
        )

        logger.debug(
            "Terminating processes..."
        )
        for process in self.processes:
            process.instance_queue.put(StopCommand)

        logger.debug(
            "Joining processes..."
        )
        for process in self.processes:
            try:
                process.join(0.5)
            except Exception as e:
                logger.exception(e)

    def __call__(self, instance) -> float:
        """
        Evaluate the likelihood of an instance according to several
        analyses, in parallel.

        Parameters
        ----------
        instance
            Some instance

        Returns
        -------
        The total log likelihood
        """
        log_likelihood = 0

        count_ = 0

        for process in self.processes:
            process.instance_queue.put(
                instance
            )

        while count_ < self.n_analyses:
            for process in self.processes:
                if process.queue.empty():
                    continue
                result = process.queue.get()

                if isinstance(
                        result,
                        Exception
                ):
                    raise result

                log_likelihood += result
                count_ += 1

        return log_likelihood



import copy
from abc import ABC, abstractmethod
from scipy.interpolate import CubicSpline
from typing import List, Dict, cast

from scipy.stats import stats

from autoconf.dictable import Dictable
from autofit.mapper.model import ModelInstance


class InterpolatorPath:
    def __init__(self, keys: List[str]):
        """
        Addresses a given attribute in a ModelInstance

        Parameters
        ----------
        keys
            A list of attribute names
        """
        self.keys = keys

    def __getattr__(self, item: str) -> "InterpolatorPath":
        """
        Add a new attribute name to the end of the path
        """
        return InterpolatorPath(self.keys + [item])

    def get_value(self, instance: ModelInstance) -> float:
        """
        Retrieve the value at the path for a given instance.

        Parameters
        ----------
        instance
            An instance of some model

        Returns
        -------
        The value for the instance at the given path.
        """
        for key in self.keys:
            instance = getattr(instance, key)
        return instance

    def __eq__(self, other: float) -> "Equality":
        """
        Create an object describing the value the addressed attribute
        should have for interpolation.

        Parameters
        ----------
        other
            A value to which the instance will be interpolated

        Returns
        -------
        An object describing how to interpolate
        """
        return Equality(self, other)


class Equality:
    def __init__(self, path: InterpolatorPath, value: float):
        """
        Describes the value of a given attribute for which other values
        are interpolated.

        Parameters
        ----------
        path
            The path to an attribute
        value
            The value that attribute should have
        """
        self.path = path
        self.value = value


class AbstractInterpolator(Dictable, ABC):
    def __init__(self, instances: List[ModelInstance]):
        """
        A TimeSeries allows interpolation on any variable.

        For example, each instance may have an attribute t. Other attributes
        can be determined for any given value of t by interpolating their
        values for each instance in the time series.

        Parameters
        ----------
        instances
            A list of instances (e.g. best fits) each with all the same attributes
            and a value on which a time series may be built
        """
        self.instances = instances

    def __getattr__(self, item: str) -> InterpolatorPath:
        """
        Used to indicate which attribute is the time attribute.

        This attribute may be at any path in the model.

        e.g.
        instance.t
        instance.something.t

        Parameters
        ----------
        item
            The name of an attribute of the instance

        Returns
        -------
        A class that keeps track of which attributes have been addressed
        """
        return InterpolatorPath([item])

    def _value_map(self, path: InterpolatorPath) -> Dict[float, ModelInstance]:
        """
        Maps know values to corresponding instances for a given path

        Parameters
        ----------
        path
            A path to an attribute, e.g. time

        Returns
        -------
        A dictionary mapping values of that attribute for each instance to the corresponding
        instance
        """
        return {path.get_value(instance): instance for instance in self.instances}

    def __getitem__(self, item: Equality) -> ModelInstance:
        """
        Create an artificial model instance which has values interpolated
        for a given interpolation value.

        Parameters
        ----------
        item
            Indicates a value for a given attribute to which the instance should
            be interpolated

        Returns
        -------
        An artificial instance with values interpolated

        Examples
        --------
        # Each instance in the time_series has an attribute 't'
        time_series = af.LinearTimeSeries([instance_1, instance_2, instance_3])

        # We can now create an instance which has a value of t = 3.5 by interpolating
        instance = time_series[time_series.t == 3.5)

        # We can also interpolate on any arbitrary attribute of the instance
        instance = time_series[time_series.some.arbitrary.attribute == -1.0]
        """
        value_map = self._value_map(item.path)
        try:
            return value_map[item.value]
        except KeyError:
            value_map = self._value_map(item.path)
            x = sorted(value_map)

            instance = self.instances[0]
            new_instance = copy.copy(instance)

            for path, _ in instance.path_instance_tuples_for_class(float):
                y = [value_map[value].object_for_path(path) for value in x]

                new_instance = new_instance.replacing_for_path(
                    path, self._interpolate(x, cast(List[float], y), item.value),
                )

        new_instance.replacing_for_path(tuple(item.path.keys), item.value)

        return new_instance

    @staticmethod
    @abstractmethod
    def _interpolate(x: List[float], y: List[float], value: float) -> float:
        """
        Interpolate a given attribute to find its effective value at some time

        Parameters
        ----------
        x
            A list of times (or another series)
        y
            A list which one value for each time
        value
            The time for which we want an interpolated value for the attribute

        Returns
        -------
        An interpolated value for the attribute
        """


class LinearInterpolator(AbstractInterpolator):
    """
    Assume all attributes have a linear relationship with time
    """

    @staticmethod
    def _interpolate(x, y, value):
        slope, intercept, r, p, std_err = stats.linregress(x, y)
        return slope * value + intercept


class SplineInterpolator(AbstractInterpolator):
    """
    Interpolate data with a piecewise cubic polynomial which is twice continuously differentiable
    """

    @staticmethod
    def _interpolate(x, y, value):
        f = CubicSpline(x, y)
        return f(value)



import configparser
import logging
import os
import random
from abc import ABC, abstractmethod
from typing import Dict, Tuple, List

import numpy as np

from autofit import exc
from autofit.mapper.prior.abstract import Prior
from autofit.mapper.prior_model.abstract import AbstractPriorModel

logger = logging.getLogger(__name__)


class AbstractInitializer(ABC):
    """
    Family of classes used to provide initial points for non-linear search
    """

    @abstractmethod
    def _generate_unit_parameter_list(self, model):
        pass

    def samples_from_model(
        self,
        total_points: int,
        model: AbstractPriorModel,
        fitness_function,
        use_prior_medians: bool = False,
        test_mode_samples: bool = True,
    ):
        """
        Generate the initial points of the non-linear search, by randomly drawing unit values from a uniform
        distribution between the ball_lower_limit and ball_upper_limit values.

        Parameters
        ----------
        total_points
            The number of points in non-linear paramemter space which initial points are created for.
        model
            An object that represents possible instances of some model with a given dimensionality which is the number
            of free dimensions of the model.
        """

        if os.environ.get("PYAUTOFIT_TEST_MODE") == "1" and test_mode_samples:
            return self.samples_in_test_mode(total_points=total_points, model=model)

        logger.info(
            "Generating initial samples of model, which are subject to prior limits and other constraints."
        )

        unit_parameter_lists = []
        parameter_lists = []
        figures_of_merit_list = []

        point_index = 0

        while point_index < total_points:
            if not use_prior_medians:
                unit_parameter_list = self._generate_unit_parameter_list(model)
            else:
                unit_parameter_list = [0.5] * model.prior_count

            parameter_list = model.vector_from_unit_vector(
                unit_vector=unit_parameter_list
            )

            try:
                figure_of_merit = fitness_function.figure_of_merit_from(
                    parameter_list=parameter_list
                )

                if np.isnan(figure_of_merit):
                    raise exc.FitException

                unit_parameter_lists.append(unit_parameter_list)
                parameter_lists.append(parameter_list)
                figures_of_merit_list.append(figure_of_merit)
                point_index += 1
            except exc.FitException:
                pass

        if total_points > 1 and np.allclose(
            a=figures_of_merit_list[0], b=figures_of_merit_list[1:]
        ):
            raise exc.InitializerException(
                """
                The initial samples all have the same figure of merit (e.g. log likelihood values).
                
                The non-linear search will therefore not progress correctly.
                
                Possible causes for this behaviour are:
                
                - The `log_likelihood_function` of the analysis class is defined incorrectly.
                - The model parameterization creates numerically inaccurate log likelihoods.
                - The`log_likelihood_function`  is always returning `nan` values.            
                """
            )

        return unit_parameter_lists, parameter_lists, figures_of_merit_list

    def samples_in_test_mode(self, total_points: int, model: AbstractPriorModel):
        """
        Generate the initial points of the non-linear search in test mode. Like normal, test model draws points, by
        randomly drawing unit values from a uniform distribution between the ball_lower_limit and ball_upper_limit
        values.

        However, the log likelihood function is bypassed and all likelihoods are returned with a value -1.0e99. This
        is so that integration testing of large-scale model-fitting projects can be performed efficiently by bypassing
        sampling of points using the `log_likelihood_function`.

        Parameters
        ----------
        total_points
            The number of points in non-linear paramemter space which initial points are created for.
        model
            An object that represents possible instances of some model with a given dimensionality which is the number
            of free dimensions of the model.
        """

        logger.warning(
            f"TEST MODE ON: SAMPLES BEING ASSIGNED ABRITRARY LARGE LIKELIHOODS"
        )

        unit_parameter_lists = []
        parameter_lists = []
        figure_of_merit_list = []

        point_index = 0

        figure_of_merit = -1.0e99

        while point_index < total_points:

            try:
                unit_parameter_list = self._generate_unit_parameter_list(model)
                parameter_list = model.vector_from_unit_vector(
                    unit_vector=unit_parameter_list
                )
                model.instance_from_vector(vector=parameter_list)
                unit_parameter_lists.append(unit_parameter_list)
                parameter_lists.append(parameter_list)
                figure_of_merit_list.append(figure_of_merit)
                figure_of_merit *= 10.0
                point_index += 1
            except exc.FitException:
                pass

        return unit_parameter_lists, parameter_lists, figure_of_merit_list


class SpecificRangeInitializer(AbstractInitializer):
    def __init__(
        self,
        parameter_dict: Dict[Prior, Tuple[float, float]],
        lower_limit=0.0,
        upper_limit=1.0,
    ):
        """
        Initializer that allows the range of possible starting points for each prior
        to be specified explicitly.

        Parameters
        ----------
        parameter_dict
            A dictionary mapping priors to inclusive ranges of physical values that
            the initial values for that dimension in the search may take
        lower_limit
            A default, unit lower limit used when a prior is not specified
        upper_limit
            A default, unit upper limit used when a prior is not specified
        """
        self.parameter_dict = parameter_dict
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    def _generate_unit_parameter_list(self, model: AbstractPriorModel) -> List[float]:
        """
        Generate a unit vector for the model. The default limits are used for any
        priors which the model has but are not found in the parameter dict.

        Parameters
        ----------
        model
            A model for which initial points are required

        Returns
        -------
        A unit vector
        """

        unit_parameter_list = []
        for prior in model.priors_ordered_by_id:
            try:
                lower, upper = map(prior.unit_value_for, self.parameter_dict[prior])
                value = random.uniform(lower, upper)
            except KeyError:
                logger.warning(
                    f"Range for {'.'.join(model.path_for_prior(prior))} not set in the SpecificRangeInitializer. "
                    f"Using defaults."
                )
                lower = self.lower_limit
                upper = self.upper_limit

                value = prior.unit_value_for(prior.random(lower, upper))

            unit_parameter_list.append(value)

        return unit_parameter_list


class Initializer(AbstractInitializer):
    def __init__(self, lower_limit: float, upper_limit: float):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        `NonLinearSearch` to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.
        """
        self.lower_limit = lower_limit
        self.upper_limit = upper_limit

    @classmethod
    def from_config(cls, config):
        """
        Load the Initializer from a non_linear config file.
        """

        try:

            initializer = config("initialize", "method")

        except configparser.NoSectionError:

            return None

        if initializer in "prior":

            return InitializerPrior()

        elif initializer in "ball":

            ball_lower_limit = config("initialize", "ball_lower_limit")
            ball_upper_limit = config("initialize", "ball_upper_limit")

            return InitializerBall(
                lower_limit=ball_lower_limit, upper_limit=ball_upper_limit
            )

    def _generate_unit_parameter_list(self, model):
        return model.random_unit_vector_within_limits(
            lower_limit=self.lower_limit, upper_limit=self.upper_limit
        )


class InitializerPrior(Initializer):
    def __init__(self):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        `NonLinearSearch` to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.

        The InitializerPrior class generates from the priors, by drawing all values as unit values between 0.0 and 1.0
        and mapping them to physical values via the prior.
        """
        super().__init__(lower_limit=0.0, upper_limit=1.0)


class InitializerBall(Initializer):
    def __init__(self, lower_limit: float, upper_limit: float):
        """
        The Initializer creates the initial set of samples in non-linear parameter space that can be passed into a
        `NonLinearSearch` to define where to begin sampling.

        Although most non-linear searches have in-built functionality to do this, some do not cope well with parameter
        resamples that are raised as FitException's. Thus, PyAutoFit uses its own initializer to bypass these problems.

        The InitializerBall class generates the samples in a small compact volume or 'ball' in parameter space, which is
        the recommended initialization strategy for the MCMC `NonLinearSearch` Emcee.

        Parameters
        ----------
        lower_limit
            The lower limit of the uniform distribution unit values are drawn from when initializing walkers in a small
            compact ball.
        upper_limit
            The upper limit of the uniform distribution unit values are drawn from when initializing walkers in a small
            compact ball.
        """
        super().__init__(lower_limit=lower_limit, upper_limit=upper_limit)



from typing import Optional

from autofit.mapper.prior_model.abstract import AbstractPriorModel
from autofit.non_linear.samples import Samples, Sample

import numpy as np


class SamplesLBFGS(Samples):

    @classmethod
    def from_results_internal(
            cls,
            results_internal: np.ndarray,
            log_posterior_list: np.ndarray,
            model: AbstractPriorModel,
            total_iterations: int,
            time: Optional[float] = None,
    ):
        """
        The `Samples` classes in **PyAutoFit** provide an interface between the results of a `NonLinearSearch` (e.g.
        as files on your hard-disk) and Python.

        To create a `Samples` object after an `pyswarms` model-fit the results must be converted from the
        native format used by `pyswarms` (which are numpy ndarrays) to lists of values, the format used by
        the **PyAutoFit** `Samples` objects.

        This classmethod performs this conversion before creating a `SamplesLBFGS` object.

        Parameters
        ----------
        results_internal
            The LBFGS results in their native internal format from which the samples are computed.
        log_posterior_list
            The log posterior of the PySwarms accepted samples.
        model
            Maps input vectors of unit parameter values to physical values and model instances via priors.
        total_iterations
            The total number of PySwarms iterations, which cannot be estimated from the sample list (which contains
            only accepted samples).
        time
            The time taken to perform the model-fit, which is passed around `Samples` objects for outputting
            information on the overall fit.
        """
        parameter_lists = [list(results_internal)]
        log_prior_list = model.log_prior_list_from(parameter_lists=parameter_lists)
        log_likelihood_list = [
            lp - prior
            for lp, prior
            in zip(log_posterior_list, log_prior_list)
        ]
        weight_list = len(log_likelihood_list) * [1.0]

        sample_list = Sample.from_lists(
            model=model,
            parameter_lists=parameter_lists,
            log_likelihood_list=log_likelihood_list,
            log_prior_list=log_prior_list,
            weight_list=weight_list
        )

        return SamplesLBFGS(
            model=model,
            sample_list=sample_list,
            total_iterations=total_iterations,
            time=time,
            results_internal=results_internal,
        )

    @property
    def x0(self):
        """
        Makes internal results accessible as `self.x0` for consistency with L-BFGS API.
        """
        return self.results_internal


import copy
from typing import List, Tuple, Union, Dict

from autofit.mapper.model_object import ModelObject
from autofit.mapper.prior_model.attribute_pair import (
    cast_collection,
    PriorNameValue,
    InstanceNameValue,
)
from .abstract import Prior

NameValue = Tuple[
    str,
    Union[Prior, float]
]


class TuplePrior(ModelObject):
    """
    A prior comprising one or more priors in a tuple
    """

    def __init__(self, **kwargs):
        super().__init__()
        for key, value in kwargs.items():
            setattr(self, key, value)

    @property
    @cast_collection(PriorNameValue)
    def prior_tuples(self):
        """
        Returns
        -------
        priors: [(String, Prior)]
            A list of priors contained in this tuple
        """
        return list(filter(lambda t: isinstance(t[1], Prior), self.__dict__.items()))

    @property
    def priors(self):
        return [prior for _, prior in self.prior_tuples]

    @property
    def unique_prior_tuples(self):
        return self.prior_tuples

    @property
    @cast_collection(InstanceNameValue)
    def instance_tuples(self):
        """
        Returns
        -------
        instances: [(String, instance)]
            A list of instances
        """
        return list(
            sorted(
                filter(lambda t: isinstance(t[1], float), self.__dict__.items()),
                key=lambda tup: tup[0],
            )
        )

    def value_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        tuple: (float,...)
            A tuple of float values
        """

        def convert(tup):
            if hasattr(tup, "prior"):
                return arguments[tup.prior]
            return tup.instance

        return tuple(
            map(
                convert,
                sorted(
                    self.prior_tuples + self.instance_tuples, key=lambda tup: tup.name
                ),
            )
        )

    def gaussian_tuple_prior_for_arguments(self, arguments):
        """
        Parameters
        ----------
        arguments: {Prior: float}
            A dictionary of arguments

        Returns
        -------
        tuple_prior: TuplePrior
            A new tuple prior with gaussian priors
        """
        tuple_prior = TuplePrior()
        for name, prior in self.prior_tuples:
            setattr(
                tuple_prior,
                name,
                arguments[prior]
            )
        for name, value in self.instance_tuples:
            setattr(
                tuple_prior,
                name,
                value
            )
        return tuple_prior

    @property
    def tuples(self) -> List[NameValue]:
        """
        The names and instances of all priors and constants ordered
        by their name.

        This means they are in the order they should be in the tuple.
        """
        return sorted(
            self.prior_tuples + self.instance_tuples,
            key=lambda t: t[0]
        )

    def _with_paths(
            self,
            tree: Dict[str, dict]
    ) -> "TuplePrior":
        """
        An instance of this tuple prior with only tuples with positions
        indicated in the tree dictionary.

        Note applying this twice will give unexpected results.
        """
        new = TuplePrior()
        for key in tree:
            key, value = self._get_key_value(key)
            setattr(new, key, value)
        return new

    def _without_paths(
            self,
            tree: Dict[str, dict]
    ) -> "TuplePrior":
        """
        An instance of this tuple prior without tuples with positions
        indicated in the tree dictionary.

        Note applying this twice will give unexpected results.
        """
        new = copy.deepcopy(self)
        for key in tree:
            key, value = self._get_key_value(key)
            delattr(new, key)
        return new

    def _get_key_value(
            self,
            key: Union[str, int]
    ) -> NameValue:
        """
        Retrieve a key and value by an attribute name or index which may
        be expressed as a string or integer.
        """
        try:
            return self.tuples[int(key)]
        except ValueError:
            return key, getattr(
                self, key
            )

    def __getitem__(self, item):
        return self._get_key_value(item)[1]



import json
import os
import sys
import zipfile
from contextlib import contextmanager
from functools import wraps
from pathlib import Path

import numpy as np

from autoconf import conf

def split_paths(func):
    """
    Split string paths if they are passed.

    e.g. "lens.mass.centre" -> ["lens", "mass", "centre"]
    """
    @wraps(func)
    def wrapper(self, paths):
        paths = [
            path.split(".")
            if isinstance(
                path, str
            ) else path
            for path in paths
        ]
        return func(self, paths)

    return wrapper


class IntervalCounter:
    def __init__(self, interval):
        self.count = 0
        self.interval = interval

    def __call__(self):
        if self.interval == -1:
            return False
        self.count += 1
        return self.count % self.interval == 0


def zip_directory(
        source_directory,
        output=None
):
    output = output or f"{source_directory}.zip"
    with zipfile.ZipFile(output, "w", zipfile.ZIP_DEFLATED) as f:
        for root, dirs, files in os.walk(source_directory):

            for file in files:
                f.write(
                    os.path.join(root, file),
                    os.path.join(
                        root[len(str(source_directory)):], file
                    ),
                )


def open_(filename, *flags):
    directory = Path(
        filename
    )
    os.makedirs(
        directory.parent,
        exist_ok=True
    )
    return open(filename, *flags)


@contextmanager
def suppress_stdout():
    with open(os.devnull, "w") as devnull:
        old_stdout = sys.stdout
        sys.stdout = devnull
        try:
            yield
        finally:
            sys.stdout = old_stdout


def numpy_array_to_json(
        array: np.ndarray, file_path: str, overwrite: bool = False
):
    """
    Write a NumPy array to a json file.

    Parameters
    ----------
    array
        The array that is written to json.
    file_path
        The full path of the file that is output, including the file name and `.json` extension.
    overwrite
        If `True` and a file already exists with the input file_path the .json file is overwritten. If 
        `False`, an error will be raised.

    Returns
    -------
    None

    Examples
    --------
    array_2d = np.ones((5,5))
    numpy_array_to_json(array_2d=array_2d, file_path='/path/to/file/filename.json', overwrite=True)
    """

    file_dir = os.path.split(file_path)[0]

    if not os.path.exists(file_dir):
        os.makedirs(file_dir)

    if overwrite and os.path.exists(file_path):
        os.remove(file_path)

    with open(file_path, "w+") as f:
        json.dump(array.tolist(), f)


def numpy_array_from_json(file_path: str):
    """
    Read a 1D NumPy array from a .json file.

    After loading the NumPy array, the array is flipped upside-down using np.flipud. This is so that the structures
    appear the same orientation as .json files loaded in DS9.

    Parameters
    ----------
    file_path
        The full path of the file that is loaded, including the file name and ``.json`` extension.

    Returns
    -------
    ndarray
        The NumPy array that is loaded from the .json file.

    Examples
    --------
    array_2d = numpy_array_from_json(file_path='/path/to/file/filename.json')
    """
    with open(file_path, "r") as f:
        return np.asarray(json.load(f))


def info_whitespace():
    return conf.instance["general"]["output"]["info_whitespace_length"]